#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """

import argparse
import json
import logging
import os.path
import serial.tools.list_ports
from os.path import join, abspath
import os
import time
from wiliot_api import ManufacturingClient
from wiliot_core import WiliotGateway, GetApiKey
from wiliot_testers.tester_utils import CsvLog, HeaderType, TesterName
from wiliot_testers.utils.upload_to_cloud_api import upload_to_cloud_api
from wiliot_testers.utils.get_version import get_version
from wiliot_testers.wiliot_tester_log import WiliotTesterLog, dict_to_csv
from wiliot_testers.wiliot_tester_tag_test import WiliotTesterTagTest
from wiliot_tools.test_equipment.test_equipment import Attenuator, BarcodeScanner, YoctoSensor
from matplotlib import pyplot as plt
import numpy as np

QR_SCANNER_TIMEOUT = 1000  # in msec to scan the external ID


class SampleChamber:

    def __init__(self, mode="offline",
                 external_id=None,
                 test=None,
                 env=None,
                 owner_id=None,
                 surface='no simulation',
                 output_path=None,
                 name=''):
        """
                Initialize the SampleChamber object.

                Parameters: - mode (str): The mode of operation, either "online" or "offline". - external_id (str):
                The external ID of the tag. - test (str): The name of the test to perform according to the tests in
                .default_test_configs.json. - env (str): The environment for cloud connection, either "prod" or
                "test". - surface (str): The surface to simulate, default is 'no simulation' - data according to the
                .default_surfaces.json. - output_path (str): The path where output files will be stored,
                if not mentioned - it will be saved in <USER>/AppData/Local/wiliot/sample/logs directory.

                Initializes various instance variables and starts the test.
                """
        mode = mode.lower() if mode else mode
        env = env.lower() if env else env
        owner_id = owner_id.lower() if owner_id else owner_id
        self.validate_args(mode, env, owner_id)
        self.run_data_path = None
        self.packets_data_path = None
        self.name = None
        self.output = output_path
        self.test_data = None
        self.runDataDict = None
        self.gw_version = None
        self.sensor = None
        self.online = True if mode.lower() == 'online' else False
        self.client = None
        self.owner_id = owner_id
        self.external_id = external_id if external_id is not None else ''
        self.common_run_name = name + time.strftime("_%d%m%y_%H%M%S")
        self.surface = surface
        self.pywiliot_version = get_version()
        self.selected_test_suite = {}
        self.barcode = None
        self.surface_freq_shift = None
        self.test_suite = {}
        self.default_dict = None
        self.test = test
        self.packets_csv = None
        upload = False

        self.logger = self.test_configs()  # Read all default values from JSONs
        self.gateway_obj = self.com_connect(self.default_dict)  # Connect to all HW
        # Init Wiliot Tester SDK
        self.wiliot_tag_test = WiliotTesterTagTest(test_suite=self.selected_test_suite,
                                                   selected_test=self.test,
                                                   logger_name=self.logger.logger.name,
                                                   logger_result_name=self.logger.results_logger.name,
                                                   logger_gw_name=self.logger.gw_logger.name,
                                                   gw_obj=self.gateway_obj)
        if self.online:
            self.connect_to_cloud()

        if self.test in ['BLE_CALIBRATION', 'BEACON_CALIBRATION', 'LORA_CALIBRATION']:
            atten_type = self.test.split('_')[0]
            self.perform_calibration_test(atten_type)

        else:
            result = self.start_test()
            if result:
                self.logger.logger.info(f'Test done, \nData stored at {self.logger.log_path}')

                if result and self.online:
                    upload = self.cloud_upload(env=env)

                if not upload:
                    if self.online:
                        self.logger.logger.warning('Failed to upload data to the cloud, please upload it manually')
                    else:
                        self.logger.logger.warning('Please upload data manually to cloud to complete post process')

        self.gateway_obj.exit_gw_api()

    @staticmethod
    def validate_args(mode, env, owner_id):
        """
                Validate the arguments passed to the constructor.

                Parameters:
                - mode (str): The mode of operation, either "online" or "offline".
                - env (str): The environment for cloud connection.
                - external_id (str): The external ID of the tag.

                Raises ValueError if the arguments are not valid.
                """
        if mode == "online" and ((not env) or (not owner_id)):
            raise ValueError("In 'online' mode - env and owner_id is mandatory")

    def test_configs(self):
        """
        Configure the test environment.

        Reads configurations from JSON files and validates them.
        Also sets up logging and test suites according to the test provided.
        """

        def convert_test_suite():
            """
            Converting the test from .default_test_configs.json to test suite that
            will fit to Wiliot Tester SDK pattern
            """
            return {
                self.test: {
                    "plDelay": 0,
                    "rssiThresholdHW": int(self.default_dict.get('rssiThresholdHW', 0)),
                    "rssiThresholdSW": int(self.default_dict.get('rssiThresholdSW', 9999)),
                    "tests": [{
                        "name": 'Sample_Test',
                        "rxChannel": self.test_suite[self.test].get('channel'),
                        "energizingPattern": self.test_suite[self.test].get('pattern'),
                        "timeProfile": [int(self.test_suite[self.test].get('tOn')),
                                        int(self.test_suite[self.test].get('tTotal'))],
                        "absGwTxPowerIndex": -1,
                        "maxTime": self.test_suite[self.test].get('testTime'),
                        "delayBeforeNextTest": int(self.test_suite[self.test].get('sleep')),
                        "stop_criteria": {'tbp_mean': [0, 9999]},
                        "quality_param": {}
                    }]
                }
            }

        def convert_calibration_suite():
            return {
                self.test: {
                    "plDelay": 0,
                    "rssiThresholdHW": int(self.default_dict.get('rssiThresholdHW', 0)),
                    "rssiThresholdSW": int(self.default_dict.get('rssiThresholdSW', 9999)),
                    "tests": [{
                        "name": 'Calibration',
                        "rxChannel": 37,
                        "energizingPattern": self.test_suite[self.test].get('pattern'),
                        "timeProfile": [int(self.test_suite[self.test].get('tOn')),
                                        int(self.test_suite[self.test].get('tTotal'))],
                        "absGwTxPowerIndex": -1,
                        "maxTime": self.test_suite[self.test].get('testTime'),
                        "delayBeforeNextTest": int(self.test_suite[self.test].get('sleep')),
                        "stop_criteria": {'tbp_mean': [0, 9999]},
                        "quality_param": {}
                    }]
                }
            }

        logger = None
        try:

            timestamp = time.strftime("_%d%m%y_%H%M%S")
            name = str(self.external_id) + '_' + str(timestamp) if self.external_id else str(timestamp)
            logger = WiliotTesterLog(run_name=name)
            self.run_data_path, self.packets_data_path = logger.create_data_dir(self.output,
                                                                                tester_name='SampleChamber',
                                                                                run_name=name)
            logger.set_logger(log_path=self.output, tester_name='SampleChamber')
            logger.set_console_handler_level(logger.logger, logging.DEBUG)
            logger.set_console_handler_level(logger.gw_logger, logging.WARNING)
            logger.set_console_handler_level(logger.results_logger, logging.DEBUG)
            sample_test_dir = os.getcwd()
            config_dir = abspath(join(sample_test_dir, 'configs'))

            with open(abspath(join(config_dir, '.defaults.json')), 'r') as defaultComs:
                self.default_dict = json.load(defaultComs)

            with open(abspath(join(config_dir, '.default_test_configs.json')), 'r') as defaultTestConfig:
                self.test_suite = json.load(defaultTestConfig)
                if self.test in ['BLE_CALIBRATION', 'BEACON_CALIBRATION', 'LORA_CALIBRATION']:
                    self.selected_test_suite = convert_calibration_suite()
                else:
                    self.selected_test_suite = convert_test_suite()

            with open(abspath(join(config_dir, '.default_surfaces.json')), 'r') as defaultSurfaces:
                surface_configs = json.load(defaultSurfaces)
                self.default_freq = surface_configs['EmulateSurface'].get('no simulation', {})
                self.surface_freq_shift = surface_configs['EmulateSurface'].get(self.surface, {})

        except Exception as e:
            if logger is not None:
                logger.logger.error(f'Could not read configs: {e}')
            raise Exception(e)

        return logger

    def com_connect(self, default_coms=None):
        """
                Connect to all hardware components.

                Parameters:
                - default_coms (dict): Dictionary containing default COM port configurations.

                Connects to the Gateway, Attenuators [BLE and LORA],
                Barcode Scanner (if available), and Temperature Sensors (if available).
                """
        self.default_dict = default_coms
        gateway_obj = None
        try:
            gw_comport = self.default_dict.get('gw')
            atten_comport = self.default_dict.get('atten')

            # Connect to the GW
            gateway_obj = WiliotGateway(auto_connect=False, port=gw_comport,
                                        logger_name=self.logger.gw_logger.name)
            if gateway_obj.is_connected():
                self.gw_version = gateway_obj.get_gw_version()
                self.logger.logger.info('GW connected')
                if self.surface_freq_shift != self.default_freq:
                    gateway_obj.write('!set_sub_1_ghz_energizing_frequency {}'.format(self.surface_freq_shift),
                                      with_ack=True)
            else:
                raise Exception('GW is not connected')

            # Connect to attenuators
            if atten_comport.get('Ble') != '':
                self.ble_atten = Attenuator('API', comport=atten_comport.get('Ble')).GetActiveTE()

                if self.ble_atten.is_open():
                    self.ble_atten.Setattn(
                        float(self.test_suite[self.test].get('attenBle', 0)))  # Line to set attenuation
                    self.logger.logger.info(f'BLE Attenuator connected - Attenuation is {self.ble_atten.Getattn()}')

                else:
                    raise Exception('BLE Attenuator not connected')

            if atten_comport.get('LoRa') != '':
                self.lora_atten = Attenuator('API', comport=atten_comport.get('LoRa')).GetActiveTE()

                if self.lora_atten.is_open():
                    self.lora_atten.Setattn(
                        float(self.test_suite[self.test].get('attenLoRa', 0)))  # Line to set attenuation
                    self.logger.logger.info(f'LORA Attenuator connected - Attenuation is {self.lora_atten.Getattn()}')

                else:
                    raise Exception('LORA Attenuator not connected')

            # Connect to barcode if needed
            if not self.external_id and self.online:
                try:
                    barcode_comport = self.default_dict.get('barcodes')
                    self.barcode = BarcodeScanner(com_port=barcode_comport, timeout=str(QR_SCANNER_TIMEOUT))
                    if not self.barcode.is_open():
                        try:
                            self.barcode.open_port(com_port=self.default_dict.get('barcodes'),
                                                   timeout=str(QR_SCANNER_TIMEOUT))
                        except Exception as e:
                            raise Exception(f'Could not connect to barcode scanner and external ID was not passed {e}')

                    self.external_id, _, _, _ = self.barcode.scan_ext_id()

                    if not self.external_id:
                        raise Exception()

                except Exception as e:
                    self.logger.logger.error(
                        f'Unable to connect to barcode scanner, need external id or scanner to scan the tag {e}')

            # Connect to sensor if available
            sensor_comport = self.default_dict.get('temperature_sensors', '')
            if len(sensor_comport) > 0:
                self.sensor = YoctoSensor(logger=self.logger.logger.name)
                self.sensor.connect()
                self.logger.logger.info('LORA Attenuator connected')

        except Exception as e:
            available_ports = [port.description for port in serial.tools.list_ports.comports()]
            self.logger.logger.error(
                f'Please redefine the comports in defaults.json, \nAvailable comports: {available_ports}, {e}')
            exit()

        return gateway_obj

    def connect_to_cloud(self):
        """
        Establishes a secure connection to the cloud service for data storage and validation.

        Parameters: - env (str): Specifies the cloud environment to connect to. It can be either 'prod' for
        production or 'test' for testing.

        Steps:
        1. Reads the user's cloud configuration file to obtain API keys and owner IDs.
        2. Validates the owner ID against a predefined list of valid owner IDs.
        3. Initializes the cloud client using the API key and specified environment.
        4. Logs the successful establishment of the cloud connection.

        Note: If the connection fails or configurations are invalid, the function will raise an exception and close
        the Gateway port.
        """
        try:
            env = 'prod'
            g = GetApiKey(gui_type='ttk', env=env, owner_id=self.owner_id)
            api_key = g.get_api_key()
            if not api_key:
                raise Exception(f'Could not found an api key for owner id {self.owner_id} and env {env}')

            self.client = ManufacturingClient(api_key=api_key, env=env, logger_=self.logger.logger.name)
            self.logger.logger.info('Connection to the cloud was established')

        except Exception as e:
            self.close_gw_port()
            raise Exception(f'Problem connecting to cloud {e}')

    def close_gw_port(self):
        """
                Close the connection to the Gateway.

                If the Gateway is connected, this function will close the port.
                """
        self.gateway_obj.exit_gw_api()

    def perform_calibration_test(self, atten_type):
        calibration_data = {}
        end_calib = False
        if atten_type == 'LORA' or atten_type == 'BEACON':
            self.gateway_obj.write('!sub1g_sync 1', with_ack=True)
            self.wiliot_tag_test.selected_test['tests'][0]['sub1g_power'] = 29
        selected_tag = None
        for atten in range(0, 31):  # Sweep from 0 to 30
            self.set_attenuation(atten_type, atten)
            res = self.wiliot_tag_test.run()
            if not res.is_results_empty():
                list_of_tags = list(res.tests[0].all_tags.tags.keys())
                if not selected_tag:
                    selected_tag = res.tests[0].selected_tag[0].packet_data['adv_address']
                for tag in list_of_tags:
                    if tag != selected_tag:
                        self.wiliot_tag_test.black_list.append(tag)
                        self.wiliot_tag_test.add_to_blacklist([tag])
                tbp = res.tests[0].selected_tag.get_statistics().get('tbp_mean', 0)
                if int(tbp) > 0:
                    end_calib = False
                    calibration_data[atten] = tbp
                else:
                    if end_calib:
                        break
                    end_calib = True
            else:
                if end_calib:
                    break
                end_calib = True
        self.generate_graph(calibration_data, f'{atten_type} Calibration')

    def set_attenuation(self, atten_type, atten_value):
        if atten_type == 'BLE' or atten_type == 'BEACON':
            self.ble_atten.Setattn(atten_value)
            self.logger.results_logger.info(f'BLE Atten set to {self.ble_atten.Getattn()}')
        elif atten_type == 'LORA':
            self.lora_atten.Setattn(atten_value)
            self.logger.results_logger.info(f'LORA Atten set to {self.lora_atten.Getattn()}')
        else:
            raise ValueError(f'Unknown attenuation type: {atten_type}')

    def find_slope_point(self, x, y):
        """
        Find the point where the slope is approximately equal to 1.
        """
        slopes = np.diff(y) / np.diff(x)
        closest_index = np.argmin(np.abs(slopes - 1))
        return x[closest_index], y[closest_index]

    def find_saturation_point(self, y, window_size=5, threshold=0.01):
        """
        Find the saturation point in the calibration curve.
        """
        y_diff = np.abs(np.diff(y, n=1))
        rolling_avg_change = np.convolve(y_diff, np.ones(window_size) / window_size, mode='valid')
        saturation_index = np.argmax(rolling_avg_change < threshold)
        return saturation_index + window_size // 2

    def generate_graph(self, data, title):

        plt.figure(figsize=(10, 6))
        x, y = zip(*data.items())
        plt.plot(x, y, marker='o', linestyle='-')

        if title == 'BLE Calibration' or title == 'LORA Calibration':
            slope_point = self.find_slope_point(x, y)
            plt.plot(slope_point[0], slope_point[1], 'ro')  # Red dot at the slope point
        elif title == 'BEACON Calibration':
            saturation_point = self.find_saturation_point(y)
            plt.plot(x[saturation_point], y[saturation_point], 'ro')  # Red dot at the saturation point

        plt.title(title)
        plt.xlabel('Attenuation')
        plt.ylabel('TBP Value')
        plt.grid(True)
        plt.show()

    def start_test(self):
        """
        Initiates and executes the test sequence for Wiliot pixie.

        This function performs several key tasks:
        1. Initializes the data structures for storing test results in CSV format.
        2. Executes the test by calling the `run()` method on the `wiliot_tag_test` object.
        3. Collects statistics and summary information about the test.
        4. Populates the run data dictionary with test results and statistics.
        5. Updates the packets data CSV with packet-level information.
        6. If in 'online' mode, it also validates the test results by comparing them with cloud data.

        Returns:
        - True: If the test and optional validation are successful.
        - False: If the test or validation fails.

        Note: This function will also close the Gateway port in case of an exception.
        """
        try:
            self.initialize_csv_data()  # Init the run result dictionary
            res = self.wiliot_tag_test.run()  # Begin the run

            if res.is_results_empty():
                #   Failed
                self.logger.logger.error('Test failed')
                self.runDataDict['tested'] = 0
                self.runDataDict['responded'] = 0
                self.runDataDict['passed'] = 0
                raise Exception('something went wrong during test and test was not preformed')

            selected_tag_statistics = res.tests[0].selected_tag.get_statistics()  # Get all the statistics
            test_sum = res.tests[0].get_summary()

            self.runDataDict['tested'] = 1
            self.runDataDict['testTime'] = str(res.get_total_test_duration())
            self.runDataDict['passed'] = int(res.is_all_tests_passed())
            self.runDataDict['responded'] = int(1 if res.tests[0].all_packets.__len__() > 0 else 0)
            self.runDataDict['responding[%]'] = str(
                self.runDataDict['responded'] / self.runDataDict['tested'] * 100) + '%'
            self.runDataDict['passed[%]'] = str(self.runDataDict['passed'] / self.runDataDict['tested'] * 100) + '%'

            if self.runDataDict['passed']:
                self.logger.logger.info('Test passed')
                selected_tag_first_payload = res.tests[0].selected_tag[0].get_payload()
            elif self.runDataDict['responded']:
                selected_tag_statistics = res.tests[0].all_packets.get_statistics()
                selected_tag_first_payload = res.tests[0].all_packets[0].get_payload()
                self.logger.logger.info('Test failed, tag responded but tbp could not be calculated')
            else:
                self.logger.logger.info('Test failed, tag did not respond')

            # save files:
            self.populate_run_data(selected_tag_statistics, test_sum)  # Fill run result dictionary
            self.update_packets_data(res, selected_tag_statistics)

            # Compare if online
            if self.online and self.runDataDict['responded']:
                validation = self.compare_tag(selected_tag_first_payload)
            else:
                validation = True
                self.logger.results_logger.info('Validation pass')

            if not validation:
                self.logger.logger.error('Validation went wrong')

            self.wiliot_tag_test.exit_tag_test()
            return validation

        except Exception as e:
            self.close_gw_port()
            self.logger.logger.error(f'Problem starting the test {e}')

    def compare_tag(self, payload):
        """
                Compare the tag data with cloud data.

                Parameters:
                - payload (str): The payload to compare.

                Returns True if the comparison is successful, otherwise returns False.
                """
        try:
            cloud_data = self.client.resolve_payload(payload=payload,
                                                     owner_id=self.owner_id)

            selected_tag_external_id = cloud_data['externalId']
            self.logger.results_logger.info(
                f'External_ID from cloud: {selected_tag_external_id}\nExternal_ID to compare is {self.external_id}')
            return self.external_id.upper() == selected_tag_external_id.upper()

        except Exception as e:
            self.close_gw_port()
            self.logger.logger.error(f'Problem comparing {e}')

    def initialize_csv_data(self):
        """
                Initialize the CSV data structures.

                Sets up dictionaries for run data and packet data.
                """

        test_values = self.selected_test_suite[self.test]['tests'][0]
        self.common_run_name = self.external_id + self.common_run_name if self.external_id is not None else self.common_run_name

        self.runDataDict = {
            'testerStationName': 'WiliotSample',
            'commonRunName': self.common_run_name,
            'batchName': '',
            'testerType': 'sample',
            'comments': '',
            'errors': '',
            'timeProfile': test_values.get('timeProfile', ''),
            'txPower': test_values.get('absGwTxPowerIndex', ''),
            'energizingPattern': test_values.get('energizingPattern', ''),
            'tested': '0',
            'passed': '0',
            'yield': '0',
            'inlay': self.default_dict['inlay'][0],
            'responded': '0',
            'responding[%]': '0',
            'passed[%]': '0',
            'testStatus': 'False',
            'operator': 'Wiliot',
            'testTime': '',
            'runStartTime': '',
            'runEndTime': '',
            'antennaType': 'TIKI',
            'surface': self.surface,
            'numChambers': '1',
            'gwVersion': self.gw_version,
            'pyWiliotVersion': self.pywiliot_version,
            'bleAttenuation': self.default_dict['atten']['Ble'],
            'loraAttenuation': self.default_dict['atten']['LoRa'],
            'testTimeProfilePeriod': test_values.get('timeProfile'[1], ''),
            'testTimeProfileOnTime': test_values.get('timeProfile'[0], ''),
            'ttfpAvg': '',
            'tbpAvg': '',
            'tbpStd': '',
            'rssiAvg': '',
            'maxTtfp': '',
            'controlLimits': self.default_dict.get('controlLimits', ''),
            'hwVersion': '',
            'sub1gFrequency': self.surface_freq_shift,
            'failBinStr': '',
            'failBin': ''
        }

        self.test_data = {
            'commonRunName': self.common_run_name,
            'encryptedPacket': '',
            'time': '',
            'externalId': self.external_id,
            'reel': self.common_run_name,
            'ttfp': '',
            'tbp': '',
            'adv_address': '',
            'state(tbp_exists:0,no_tbp:-1,no_ttfp:-2,dup_adv_address:-3)': '',
            'status': '',
            'chamber': '',
            'temperature_from_sensor': ''}

        dict_to_csv(dict_in=self.runDataDict, path=self.run_data_path, append=False, only_titles=True)
        self.packets_csv = CsvLog(header_type=HeaderType.PACKETS, path=self.packets_data_path,
                                  tester_type=TesterName.SAMPLE)
        self.packets_csv.open_csv()

    def populate_run_data(self, selected_tag_statistics, test_sum):
        """
                Populate the run data dictionary.

                Parameters:
                - selected_tag_statistics (dict): Statistics of the selected tag.
                - test_sum (dict): Summary of the test.

                Fills the run data dictionary with test results and statistics.
                """
        self.runDataDict['testStatus'] = bool(test_sum['is_test_pass'])
        self.runDataDict['runStartTime'] = test_sum['test_start_time'].strftime("%Y-%m-%d %H:%M:%S.%f")
        self.runDataDict['runEndTime'] = test_sum['test_end_time'].strftime("%Y-%m-%d %H:%M:%S.%f")
        self.runDataDict['ttfpAvg'] = str(selected_tag_statistics.get('ttfp', 'nan'))
        self.runDataDict['tbpAvg'] = str(selected_tag_statistics.get('tbp_mean', 'nan'))
        self.runDataDict['tbpStd'] = str(selected_tag_statistics.get('tbp_std', 'nan'))
        self.runDataDict['rssiAvg'] = str(selected_tag_statistics.get('rssi_mean', 'nan'))
        self.runDataDict['maxTtfp'] = str(selected_tag_statistics.get('ttfp', 'nan'))
        dict_to_csv(dict_in=self.runDataDict, path=self.run_data_path, append=True, only_titles=False)

    def update_packets_data(self, res=None, stats=None):
        """
                Update the packets data CSV.

                Parameters:
                - res (object): The result object containing test data.
                - stats (dict): Statistics of the selected tag.

                Updates the packets data CSV with packet information.
                """

        if res:
            r = res.tests[0]
            if r.selected_tag.__len__() > 0:  # pass
                tag_data = r.selected_tag
            elif r.all_packets.__len__() > 0:  # responded but did not pass
                tag_data = r.all_packets
            else:  # No response
                self.test_data.update({
                    'encryptedPacket': '',
                    'time': '',
                    'chamber': 0,
                    'status': r.is_test_passed,
                    'state(tbp_exists:0,no_tbp:-1,no_ttfp:-2,dup_adv_address:-3)': -2,
                    'ttfp': str(stats.get('ttfp', 'nan')),
                    'tbp': str(stats.get('tbp_min', 'nan')),
                    'adv_address': '',
                    'temperature_from_sensor': self.sensor.get_temperature()
                    if self.sensor is not None else 'nan'})
                return

            for p in tag_data:
                for i in range(p.gw_data['gw_packet'].size):
                    if p.gw_data["time_from_start"].size > 1:
                        time_from_start = p.gw_data["time_from_start"][i]
                    else:
                        time_from_start = p.gw_data["time_from_start"].item()
                    self.test_data.update({
                        'encryptedPacket': p.get_packet_content(get_gw_data=True, sprinkler_num=i),
                        'time': time_from_start,
                        'chamber': 0,
                        'status': r.is_test_passed,
                        'state(tbp_exists:0,no_tbp:-1,no_ttfp:-2,dup_adv_address:-3)': 0 if stats[
                            'tbp_mean'] else -1,
                        'ttfp': str(stats.get('ttfp', 'nan')),
                        'tbp': str(stats.get('tbp_min', 'nan')),
                        'adv_address': res.get_test_unique_adva(),
                        'temperature_from_sensor': self.sensor.get_temperature()
                        if self.sensor is not None else 'nan'})
                    self.packets_csv.append_dict_as_row([self.test_data])

    def cloud_upload(self, env):
        """
        Uploads the collected test data to the cloud for further analysis and storage.

        Parameters: - env (str): Specifies the cloud environment to upload to. It can be either 'prod' for production
        or 'test' for testing.

        The function performs the following tasks:
        1. Calls the `upload_to_cloud_api` function to upload both run data and packets data CSV files.
        2. Logs the status of the upload operation.

        Returns:
        - True: If the data upload to the cloud is successful.
        - False: If the data upload fails for any reason.

        Note: The function uses the owner ID and API key that were set during the cloud connection phase.
        """
        success = upload_to_cloud_api(batch_name='',
                                      tester_type='sample-test',
                                      run_data_csv_name=self.run_data_path,
                                      packets_data_csv_name=self.packets_data_path,
                                      is_path=True,
                                      env=env,
                                      owner_id=self.owner_id,
                                      logger_=self.logger.logger.name)

        return success


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Sample Chamber")

    parser.add_argument('-n', '--name', default="", help="The test name (common run name prefix)")
    parser.add_argument('-m', '--mode', default="offline", choices=['online', 'offline'],
                        help="Online mode when using cloud, Offline with barcode scanner")
    parser.add_argument('-o', '--ownerid',
                        help="Owner ID for getting api_key from the cloud - required in online mode")
    parser.add_argument('-e', '--externalID', help="Tag External ID")
    parser.add_argument('-t', '--test', required=True, choices=['TIKI_LORA', 'TIKI_BLE',
                                                                'BLE_CALIBRATION', 'BEACON_CALIBRATION',
                                                                'LORA_CALIBRATION'],
                        help="Test to perform")
    parser.add_argument('--env', choices=['prod', 'test'], help="Environment from cloud")
    parser.add_argument('-s', '--surface', default="no simulation",
                        choices=['simulate cardboard', 'simulated RPC', 'no simulation'], help="Surface to simulate")
    parser.add_argument('--output', help="Files output")

    args = parser.parse_args()

    chamber = SampleChamber(mode=args.mode, external_id=args.externalID, test=args.test, env=args.env,
                            owner_id=args.ownerid, surface=args.surface, output_path=args.output,
                            name=args.name)

    # Examples for console:
    # python sample_chamber.py --mode offline --test TIKI_BLE --externalID (01)00850027865010(21)04z3T0049
    # python sample_chamber.py --mode online --test TIKI_BLE --externalID (01)00850027865010(21)04z3T0049 --env prod

    # Examples to execute from code:
    # mode = 'Online'
    # externalID = ''
    # test = 'TIKI_LORA'
    # env = 'prod'
    # owner_id = 'wiliot-ops'
    # surface = 'simulate cardboard'
    # chamber = SampleChamber(mode=mode, external_id=externalID, test=test, env=env, owner_id=owner_id, surface=surface)



"""
  Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

  Redistribution and use of the Software in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

     1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     2. Redistributions in binary form, except as used in conjunction with
     Wiliot's Pixel in a product or a Software update for such product, must reproduce
     the above copyright notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

     3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
     may be used to endorse or promote products or services derived from this Software,
     without specific prior written permission.

     4. This Software, with or without modification, must only be used in conjunction
     with Wiliot's Pixel or with Wiliot's cloud service.

     5. If any Software is provided in binary form under this license, you must not
     do any of the following:
     (a) modify, adapt, translate, or create a derivative work of the Software; or
     (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
     discover the source code or non-literal aspects (such as the underlying structure,
     sequence, organization, ideas, or algorithms) of the Software.

     6. If you create a derivative work and/or improvement of any Software, you hereby
     irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
     royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
     right and license to reproduce, use, make, have made, import, distribute, sell,
     offer for sale, create derivative works of, modify, translate, publicly perform
     and display, and otherwise commercially exploit such derivative works and improvements
     (as applicable) in conjunction with Wiliot's products and services.

     7. You represent and warrant that you are not a resident of (and will not use the
     Software in) a country that the U.S. government has embargoed for use of the Software,
     nor are you named on the U.S. Treasury Department’s list of Specially Designated
     Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
     You must not transfer, export, re-export, import, re-import or divert the Software
     in violation of any export or re-export control laws and regulations (such as the
     United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
     and use restrictions, all as then in effect

   THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
   OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
   WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
   QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
   IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
   ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
   FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
   (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
   (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
   (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
   (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""
import datetime
import pandas as pd
from enum import Enum

from wiliot_core import Packet, PacketList, TagCollection
try:
    from wiliot_core import DecryptedTagCollection, DecryptedPacketList, DecryptedPacket
    print('tag result: decryption mode is supported if needed')
except Exception as e:
    pass


class FailureCodes(Enum):
    NONE = 0
    PASS = 1  # Pass - default bin when no failure observed
    NO_RESPONSE = 2  # No response - no packets received at all in this location
    NO_PACKETS_UNDER_RSSI_THR = 3  # No packets under RSSI threshold
    MISSING_LABEL = 4  # Missing label - failure due to missing label
    FIRST_TAG_IN_RUN = 5  # First tag in run
    HIGH_TTFP = 6  # Quality Test Failure: High TTFP - time to first packet did not satisfy the configuration file
    NOT_ENOUGH_PACKETS = 7  # Quality Test Failure: Not enough packets - did not reach packet threshold as defined in configuration file
    TBP_AVG_OUT_OF_BOUNDS = 8  # Quality Test Failure: TBP Avg out of bounds - TBP Average value is out of bounds
    TBP_MIN_OUT_OF_BOUNDS = 9  # Quality Test Failure: TBP Min out of bounds - TBP Min value is out of bounds
    TBP_MAX_OUT_OF_BOUNDS = 10  # Quality Test Failure: TBP Max out of bounds - TBP Max value is out of bounds
    NO_TBP = 11  # Quality Test Failure: Could not calculate TBP - could not calculate TBP statistic at all
    NOT_ENOUGH_PACKETS_IN_SPRINKLER = 12  # Quality Test Failure: Not enough packets in sprinkler - received packet count in sprinkler is lower than specified in configuration file
    TIME_BTWN_CYCLES_OUT_OF_BOUNDS = 13  # Quality Test Failure: Time between cycles out of bounds
    DUPLICATION_OFFLINE = 14  # Duplicate offline - tag was found as duplicate during offline tester run (ADVA duplicate)
    DUPLICATION_POST_PROCESS = 15  # Duplicate post process - tag was found as duplicate in post process (UID duplicate)
    GW_ERROR = 16  # GW error
    STOP_BY_USER = 17  # test stopped by the user
    SEVERAL_TAGS_UNDER_TEST = 18  # Cannot decided which tag is under test
    FAILED_QUALITY_TEST = 19  # Failed Quality Test according to the specified statistics limits.
    NOT_PRINTED = 20  # tag was not printed due to printing offset
    CORRUPTED_PACKET_POST_PROCESS = 21  # post process - corrupted packet detection
    DIFF_TAG_ID_WITH_SAME_ADVA_POST_PROCESS = 22  # post process - same adva but different tag id
    ALTERING_ADVA_POST_PROCESS = 23  # post process - tag id with "good" packets sends diff adva
    BAD_PRINTING = 24  # as a result of printing validation, tag was decided to have a bad printing label
    INVALID_EXTERNAL_ID_POST_PROCESS = 25  # post process - specified if external id is empty or invalid
    WAFER_SORT_FAIL_POST_PROCESS = 26  # post process - tag id has failed in wafer sort
    END_OF_TEST = 27    # End of test for tags that was moved over the coupler for validating tested tags
    WRONG_POWER_MODE_POST_PROCESS = 28 # post process - tag has the wrong power mode based on the inlay expected power mode
    SHIPMENT_INTERNAL_ID_DUPLICATION = 90  # shipment API error - decided over the cloud
    SHIPMENT_EXTERNAL_ID_DUPLICATION = 91  # shipment API error - decided over the cloud
    SHIPMENT_TAG_NOT_EXIST = 92  # shipment API error - decided over the cloud
    SHIPMENT_OWNER_NOT_WILIOT_OPS = 93  # shipment API error - decided over the cloud
    SHIPMENT_SERIALIZATION_ISSUE = 94  # shipment API error - decided over the cloud
    SOFTWARE_GENERAL_ERROR = 999  # usage for debugging


class FailBinSample(Enum):
    TBP_EXISTS = 0
    NO_TBP = -1
    NO_TTFP = -2
    ADVA_DUPLICATION = -3
    NO_SERIALIZATION = -4
    UNDETERMINED_ERROR = -5


class ConversionTypes(Enum):
    NOT_CONVERTED = 'Unconverted'
    STANDARD = 'Regular Conversion'
    DURABLE = 'Durable Conversion'


class SurfaceTypes(Enum):
    AIR = 'air'
    CARDBOARD = 'cardboard'
    RPC = 'RPC'
    ER3 = 'Er3'
    ER3_5 = 'Er3.5'
    ER5 = 'Er5'
    ER7 = 'Er7'
    ER12 = 'Er12'


class WiliotTesterTagResultList(list):
    def __init__(self):
        self.tests = []
        self.trigger_time = None
        self.status = None
        self.test_info = {}
    
    def __len__(self):
        return len(self.tests)
    
    def __iter__(self):
        self.n = 0
        return iter(self.tests)
    
    def __next__(self):
        if self.n <= len(self):
            return self.tests[self.n]
        else:
            raise StopIteration
    
    def append(self, res) -> None:
        self.tests.append(res)
    
    def is_results_empty(self):
        """
        true if no test was done
        :return:
        :rtype:
        """
        return len(self.tests) == 0
    
    def is_all_tests_passed(self):
        """
        check if all sub test passed (i.e. the whole test passed)
        :return:
        :rtype:
        """
        if self.is_results_empty():
            return None
        return self.tests[-1].is_test_passed if self.status is None else self.status

    def get_total_test_duration(self):
        if self.is_results_empty():
            return None
        return (self.tests[-1].test_end - self.trigger_time).total_seconds()
    
    def get_total_fail_bin(self, as_name=False):
        """
        return the test status according to the s option
        :param as_name: if true, the code is return b ts name and not its value
        :type as_name: bool
        :return: the test status code according to the FailureCodes
        :rtype: int or str
        """
        if self.is_results_empty():
            return None
        if as_name:
            return self.tests[-1].test_status.name
        else:
            return self.tests[-1].test_status.value
    
    def set_total_fail_bin(self, fail_code, overwrite=False):
        """
        set the test status to the specified fail_code
        :param fail_code: the status code
        :type fail_code: FailureCodes
        :param overwrite: if true fail bin will be overwrite according to the specified fail bin
        :type overwrite: bool
        """
        if self.is_results_empty():
            return None
        if self.tests[-1].test_status == FailureCodes.NONE or self.tests[-1].test_status == FailureCodes.PASS or \
                overwrite:
            self.tests[-1].test_status = fail_code
            self.status = fail_code == FailureCodes.PASS
    
    def check_and_get_selected_tag_id(self):
        """
        check the for all sub test the same tag was selected and return the selected tag id
        :return: the selected tag id, empty string for unselected tag id
        :rtype: str
        """
        selected_tag = ''
        for r in self.tests:
            if r.selected_tag.size():
                cur_selected_tag = r.selected_tag[0].packet_data['adv_address']
                if selected_tag == '':
                    selected_tag = cur_selected_tag
                elif selected_tag != cur_selected_tag:
                    self.set_total_fail_bin(FailureCodes.SEVERAL_TAGS_UNDER_TEST)
                    self.status = False
        return selected_tag
    
    def set_trigger_time(self):
        self.trigger_time = datetime.datetime.now()
    
    def get_trigger_time(self):
        return self.trigger_time
    
    def set_total_test_status(self, status=False):
        """
        set the test pass/fail status
        :param status: if true the test was passed
        :type status: bool
        """
        self.status = status
    
    def get_total_test_status(self):
        """
        get the test pass/fail status
        :return: if true the test was passed
        :rtype: bool
        """
        if self.status is None:
            return self.is_all_tests_passed()
        else:
            return self.status
    
    def set_packet_status(self, adv_address='', status=''):
        """
        change the status per packets from outside function of the all_packets variable.
        the status can be change only if the previous status was 'good'
        :param adv_address: the packet advertising address
        :type adv_address: str
        :param status: the desired status to change
        :type status: str
        """
        for test in self.tests:
            for p in test.all_packets:
                if p.packet_data['adv_address'] == adv_address:
                    for i, s in enumerate(p.custom_data['packet_status']):
                        if s == 'good':
                            p.custom_data['packet_status'][i] = status

    def get_total_ttfp(self):
        if self.is_results_empty() or 'ttfp' not in self.tests[0].selected_tag_statistics.keys():
            return float('nan')
        return self.tests[0].selected_tag_statistics['ttfp']

    def get_test_unique_adva(self):
        all_adva = []
        for test in self.tests:
            all_adva += list(test.all_tags.tags.keys())
        return list(set(all_adva))

    def set_test_info(self, test_info):
        """
        add meta data to test
        @param test_info:
        @type test_info: dict
        @return:
        @rtype:
        """
        for k, v in test_info.items():
            self.test_info[k] = v

    def get_test_info(self):
        return self.test_info


class WiliotTesterTagResult(object):
    def __init__(self, dev_mode=False):
        self.is_test_passed = False
        self.test_status = FailureCodes.NONE
        self.test_start = None
        self.test_end = None
        self.selected_tag_statistics = {}
        self.run_data = {}
        self.dev_mode = dev_mode
        if self.dev_mode:
            self.all_packets = DecryptedPacketList()
            self.selected_tag = DecryptedPacketList()
            self.all_tags = DecryptedTagCollection()
            self.filtered_tags = DecryptedTagCollection()
        else:
            self.all_packets = PacketList()
            self.selected_tag = PacketList()
            self.all_tags = TagCollection()
            self.filtered_tags = TagCollection()
    
    def add_run_data(self, test_param):
        """
        add the test data according to the test_param
        :param test_param: dictionary contains the main parameters for all test
        :type test_param: dict
        :return:
        :rtype:
        """
        self.run_data['rx_channel'] = int(test_param['rxChannel'])
        self.run_data['energizing_pattern'] = test_param['energizingPattern']
        self.run_data['time_profile'] = test_param['timeProfile']

    def add_packets_to_multi_tags(self, packets_in, multi_tag_out):
        """
        add packets_in to multi tag object
        :param packets_in: the packets needed to be append to the specified multi tag object
        :type packets_in: pd.DataFrame or PacketList or Packet or DecryptedPacket or DecryptedPacketList
        :param multi_tag_out:
        :type multi_tag_out:
        :return: the modified multi tag object
        :rtype: TagCollection
        """
        
        if isinstance(packets_in, pd.DataFrame):
            new_packet_list = PacketList() if not self.dev_mode else DecryptedPacketList
            packets_in = new_packet_list.import_packet_df(packet_df=packets_in)

        # it is by design not elif but if due to the first cond.
        if isinstance(packets_in, PacketList) or isinstance(packets_in, DecryptedPacketList):
            for packet in packets_in:
                multi_tag_out.append(packet.copy())
        
        elif isinstance(packets_in, Packet) or isinstance(packets_in, DecryptedPacket):
            multi_tag_out.append(packets_in)
        
        else:  # TODO add more options
            raise Exception('WiliotTesterTagResult: add_packets does not support the type of the specified packets_in')
        
        return multi_tag_out
    
    def add_selected_tag_statistics(self, custom_data=None):
        """
        add the statistics of the selected tag for the test
        :param custom_data: if specified the custom data is added to the packets of the selected tag
        :type custom_data: dict
        """
        if custom_data is not None:
            for packet in self.selected_tag:
                packet.add_custom_data(custom_data=custom_data)
        
        self.selected_tag_statistics = self.selected_tag.get_statistics()
    
    def set_quality_test_failure(self, statistics_limits, targets_status):
        """
        if test was failed, the correct failure code is set according to the statistics_limits, targets_status
        :param statistics_limits: the valid range of the quality parameter
        :type statistics_limits: dict
        :param targets_status: list of boolean, true if passed the quality test
        :type targets_status: list
        """
        for t, k in zip(targets_status, statistics_limits.keys()):
            if not t:  # failed the test
                if k == 'ttfp':
                    self.test_status = FailureCodes.HIGH_TTFP
                elif 'tbp' in k and self.selected_tag_statistics[k] == -1:
                    self.test_status = FailureCodes.NO_TBP
                elif k == 'tbp_mean':
                    self.test_status = FailureCodes.TBP_AVG_OUT_OF_BOUNDS
                elif k == 'tbp_min':
                    self.test_status = FailureCodes.TBP_MIN_OUT_OF_BOUNDS
                elif k == 'tbp_max':
                    self.test_status = FailureCodes.TBP_MAX_OUT_OF_BOUNDS
                elif k == 'sprinkler_counter_max':
                    self.test_status = FailureCodes.NOT_ENOUGH_PACKETS_IN_SPRINKLER
                elif k == 'time_btw_cycle':
                    self.test_status = FailureCodes.TIME_BTWN_CYCLES_OUT_OF_BOUNDS
                else:
                    self.test_status = FailureCodes.FAILED_QUALITY_TEST
                break  # return after the first failure
    
    def get_summary(self):
        """
        get a dictionary of the test summary including the tag statistics and test parameters
        :return: test summary
        :rtype: dict
        """
        if self.selected_tag.size():
            selected_tag = self.selected_tag[0].packet_data['adv_address']
        else:
            selected_tag = ''
        summary = {'is_test_pass': self.is_test_passed, 'selected_tag': selected_tag,
                   'test_start_time': self.test_start, 'test_end_time': self.test_end,
                   'test_status': self.test_status.value}
        
        run_summary = self.run_data.copy()
        run_summary['time_profile'] = str(self.run_data['time_profile'])
        
        tag_stat = {'num_packets': 0, 'num_cycles': 0, 'sprinkler_counter_mean': None,
                    'sprinkler_counter_std': None, 'sprinkler_counter_min': None, 'sprinkler_counter_max': None,
                    'tbp_mean': None, 'tbp_std': None, 'tbp_min': None, 'tbp_max': None, 'tbp_num_vals': None,
                    'per_mean': None, 'per_std': None, 'rssi_mean': None, 'rssi_std': None, 'rssi_min': None,
                    'rssi_max': None, 'ttfp': None, 'rx_rate_normalized': None, 'rx_rate': None}
        for k in tag_stat.keys():
            if k in self.selected_tag_statistics.keys():
                tag_stat[k] = self.selected_tag_statistics[k]
        
        return {**summary, **run_summary, **tag_stat}



"""
  Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

  Redistribution and use of the Software in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

     1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     2. Redistributions in binary form, except as used in conjunction with
     Wiliot's Pixel in a product or a Software update for such product, must reproduce
     the above copyright notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

     3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
     may be used to endorse or promote products or services derived from this Software,
     without specific prior written permission.

     4. This Software, with or without modification, must only be used in conjunction
     with Wiliot's Pixel or with Wiliot's cloud service.

     5. If any Software is provided in binary form under this license, you must not
     do any of the following:
     (a) modify, adapt, translate, or create a derivative work of the Software; or
     (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
     discover the source code or non-literal aspects (such as the underlying structure,
     sequence, organization, ideas, or algorithms) of the Software.

     6. If you create a derivative work and/or improvement of any Software, you hereby
     irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
     royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
     right and license to reproduce, use, make, have made, import, distribute, sell,
     offer for sale, create derivative works of, modify, translate, publicly perform
     and display, and otherwise commercially exploit such derivative works and improvements
     (as applicable) in conjunction with Wiliot's products and services.

     7. You represent and warrant that you are not a resident of (and will not use the
     Software in) a country that the U.S. government has embargoed for use of the Software,
     nor are you named on the U.S. Treasury Department’s list of Specially Designated
     Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
     You must not transfer, export, re-export, import, re-import or divert the Software
     in violation of any export or re-export control laws and regulations (such as the
     United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
     and use restrictions, all as then in effect

   THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
   OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
   WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
   QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
   IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
   ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
   FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
   (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
   (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
   (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
   (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""

ASTRIX = ''.join(['*'] * 10)


# Configurations and Enums for offline_main.py
class MainDefaults:
    # GUI
    FONT_NAME = 'SansSerif'
    FONT_SIZE = 10
    GUI_UPDATE_TIME = 500  # ms
    DEFAULT_LANGUAGE = 'En'  # Default to English
    PRINTING_GUI_TYPE = ['SGTIN', 'Barcode', 'prePrint']

    # Modules handler
    MAX_QUEUE_SIZE = 100

    # calculation
    CALCULATE_INTERVAL = 10
    CALCULATE_ON = 50
    N_YIELD_SAMPLES_UNDER_THRESHOLD = 15
    MAX_TTFP_ERROR = 2  # if ttfp avg is more than MAX_TTFP_ERROR, the summary window will be red
    MAX_FILE_SIZE = 12 * 10 ** 6  # 12 Mb


# Configurations for offline_printing_and_validation.py
class PrintingAndValidationDefaults:
    # module handlers
    FIRST_TAG_RUN_LOCATION = 0  # first tag run location
    QUEUE_VALIDATION_OFFSET = 1  # queue size offset when no scanner is connected
    EVENT_WAIT_TIME = 20  # the max time to wait for an event
    MIN_WAIT_TIME = 2  # time to wait for post run validation validation if stop is set
    MAX_MAIN_STATES = 80

    # Printer and printing
    PRINTER_SOCKET_TIMEOUT = 1  # seconds
    PRINTER_ACK_TIMEOUT = 1.5  # seconds max time we allow to get ack from printer
    TIME_BTWN_PRINTER_REQUESTS = 0.25  # seconds
    FAIL_JOB_NUM = 1  # line number for fail job
    PASS_JOB_NUM = 2  # line number for pass job
    PASS_MIRROR_JOB_NUM = 3  # line number for pass mirrored job
    TAG_COUNT_SIZE = 4  # number of characters per tag count inside the external id

    # Printer Cart:
    PRINTER_90_DEG_WAIT_TIME = 1.5  # time to wait for the printer if NOT printing during movement
    PRINTER_FORWARD_MSG = 'high-to-low'
    PRINTER_BACKWARD_MSG = 'low-to-high'

    # scanner
    SCANNER_GOOD_READ_TIMEOUT = 1.5  # seconds max time we allow to validate scanner
    TRIGGER_TYPE = 5  # relevant only for Cognex, 5: Continuous (external)

    # r2r
    ARDUINO_BAUD_RATE = 1000000
    PASS_PULSE_GPIO = 1
    FAIL_PULSE_GPIO = 2
    START_STOP_GPIO = 3
    DIRECTION_GPIO = 5  # 0 regular run (left to right) 1 rewind (right to left)
    MISSING_LABEL_GPIO = 4
    R2R_PULSE_DURATION = 50  # ms
    MAX_R2R_WAIT_TIME = 90  # being on the safe side since r2r can wait max 99 seconds
    R2R_START_OFFSET = 1  # When starting run then its starting from 1 and not 0

    # Bendurance
    BENDURANCE_START_STOP_GPIO = 4
    BENDURANCE_GW_TRIGGER = 3


# Configurations for offline_tag_testing.py
class TagTestingDefaults:
    # missing label:
    TIMEOUT_FOR_MISSING_LABEL = 5
    ENABLE_MISSING_LABEL = False
    MAX_MISSING_LABEL_ENGINEERING = 30

    # unique testing modes:
    TIME_BETWEEN_TEST_PRINTING = 3  # timeout when do printing calibration with no RF testing
    POST_RUN_WAIT_TIME = 5  # relevant only if GW control R2R: timeout for post run validation validation if stop is set

    # GPIO and R2R
    R2R_PULSE_DURATION = 50  # relevant only if GW control R2R: milliseconds
    MAX_R2R_WAIT_TIME = PrintingAndValidationDefaults.MAX_R2R_WAIT_TIME
    EVENT_WAIT_TIME = 5  # relevant only if GW control R2R: time to wait till printing and validation are done
    START_STOP_GPIO = 'P0.31'  # relevant only if GW control R2R
    PASS_PULSE_GPIO = 'P0.29'  # relevant only if GW control R2R
    FAIL_PULSE_GPIO = 'P0.30'  # relevant only if GW control R2R
    MISSING_LABEL_GPIO = 'P1.00'  # relevant only if GW control R2R
    PRINTER_DIRECTION_GPIO = 'P1.09'  # relevant only if GW control R2R
    DIRECTION_GPIO = ''  # relevant only if GW control R2R - unsupported pin

    # Printer Cart
    PRINTER_SIGNAL_TIMEOUT = 1  # sec, max time to get ack from the printer cart that it completed its movement
    PRINTER_SIGNAL_SLEEP_TIME = 0.100  # sec, time between gw responses check



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """

import json
import logging
from os.path import join, abspath
import os
import time
from wiliot_api import ManufacturingClient
from wiliot_core import WiliotGateway, GetApiKey
from wiliot_testers.utils.upload_to_cloud_api import upload_to_cloud_api
from wiliot_testers.utils.get_version import get_version
from wiliot_testers.wiliot_tester_log import WiliotTesterLog, dict_to_csv
from wiliot_testers.wiliot_tester_tag_test import WiliotTesterTagTest
from wiliot_testers.wiliot_tester_tag_result import FailureCodes


class SimpleTest:

    def __init__(self,
                 owner_id='wiliot-ops',
                 output_path=None,
                 test_name=None,
                 test_time=None,
                 tester_type='sample',
                 test_env='test',
                 upload_to_cloud=False):
        """
                Initialize the SampleChamber object.

                Parameters: - mode (str): The mode of operation, either "online" or "offline". - external_id (str):
                The external ID of the tag. - test (str): The name of the test to perform according to the tests in
                .default_test_configs.json. - env (str): The environment for cloud connection, either "prod" or
                "test". - surface (str): The surface to simulate, default is 'no simulation' - data according to the
                .default_surfaces.json. - output_path (str): The path where output files will be stored,
                if not mentioned - it will be saved in <USER>/AppData/Local/wiliot/sample/logs directory.

                Initializes various instance variables and starts the test.
                """

        self.run_data_path = None
        self.packets_data_path = None
        self.name = test_name if test_name is not None else 'SimpleTest'
        self.output = output_path
        self.test_data = None
        self.runDataDict = None
        self.gw_version = None
        self.client = None
        self.tester_type = tester_type
        cur_time = time.strftime("_%d%m%y_%H%M%S")
        self.common_run_name = self.name + cur_time
        self.pywiliot_version = get_version()
        self.test = 'Simple Test'
        self.test_suite = None
        self.env = test_env
        self.owner_id = owner_id
        self.packets_csv = None

        self.logger = self.test_configs(test_time=test_time)  # Read all default values from JSONs
        self.gateway_obj = self.com_connect()  # Connect to all HW
        # Init Wiliot Tester SDK
        self.wiliot_tag_test = WiliotTesterTagTest(test_suite=self.test_suite,
                                                   selected_test=self.test,
                                                   logger_name=self.logger.logger.name,
                                                   logger_result_name=self.logger.results_logger.name,
                                                   logger_gw_name=self.logger.gw_logger.name,
                                                   gw_obj=self.gateway_obj)

        self.connect_to_cloud()

        result = self.start_test()

        if result and upload_to_cloud:
            self.logger.logger.info(f'Test done, \nData stored at {self.logger.log_path} \nStarting uploading to cloud')
            upload = self.cloud_upload()

            if not upload:
                self.logger.logger.warning('Failed to upload data to the cloud, please upload it manually')

        self.close_gw_port()

    def test_configs(self, test_time=None):
        """
        Configure the test environment.

        Reads configurations from JSON files and validates them.
        Also sets up logging and test suites according to the test provided.
        """

        logger = None
        try:

            logger = WiliotTesterLog()
            logger.set_common_run_name(reel_name='SimpleTest')
            logger.set_logger(log_path=self.output, tester_name='SimpleTest')
            self.run_data_path, self.packets_data_path = logger.create_data_dir(self.output,
                                                                                tester_name='SimpleTest',
                                                                                run_name=self.common_run_name)
            logger.set_console_handler_level(logger.logger, logging.DEBUG)
            logger.set_console_handler_level(logger.gw_logger, logging.WARNING)
            logger.set_console_handler_level(logger.results_logger, logging.DEBUG)
            sample_test_dir = os.getcwd()
            config_dir = abspath(join(sample_test_dir, 'configs'))

            with open(abspath(join(config_dir, 'test_suite_example.json')), 'r') as TestSuite:
                self.test_suite = json.load(TestSuite)
                if test_time is not None:
                    self.test_suite[self.test]['maxTtfp'] = int(test_time)
                    self.test_suite[self.test]['tests'][0]['maxTime'] = int(test_time)

        except Exception as e:
            if logger is not None:
                logger.logger.error(f'Could not read test suite: {e}')
            raise Exception(e)

        return logger

    def com_connect(self):
        """
                Connect to all hardware components.

                Parameters:
                - default_coms (dict): Dictionary containing default COM port configurations.

                Connects to the Gateway, Attenuators [BLE and LORA],
                Barcode Scanner (if available), and Temperature Sensors (if available).
                """
        gateway_obj = None
        try:
            # Connect to the GW
            gateway_obj = WiliotGateway(auto_connect=True,
                                        logger_name=self.logger.gw_logger.name)
            if gateway_obj.is_connected():
                self.gw_version = gateway_obj.get_gw_version()
                self.logger.logger.info('GW connected')

            else:
                raise Exception('GW is not connected')

        except Exception as e:
            self.logger.logger.error(
                f'Could not detect GW, {e}')
            exit()

        return gateway_obj

    def connect_to_cloud(self):
        """
        Establishes a secure connection to the cloud service for data storage and validation.

        Parameters: - env (str): Specifies the cloud environment to connect to. It can be either 'prod' for
        production or 'test' for testing.

        Steps:
        1. Reads the user's cloud configuration file to obtain API keys and owner IDs.
        2. Validates the owner ID against a predefined list of valid owner IDs.
        3. Initializes the cloud client using the API key and specified environment.
        4. Logs the successful establishment of the cloud connection.

        Note: If the connection fails or configurations are invalid, the function will raise an exception and close
        the Gateway port.
        """
        try:
            g = GetApiKey(gui_type='ttk', env=self.env, owner_id=self.owner_id)
            api_key = g.get_api_key()
            if not api_key:
                raise Exception(f'Could not found an api key for owner id {self.owner_id} and env {self.env}')

            self.client = ManufacturingClient(api_key=api_key, env=self.env, logger_=self.logger.logger.name)
            self.logger.logger.info('Connection to the cloud was established')

        except Exception as e:
            self.close_gw_port()
            raise Exception(f'Problem connecting to cloud {e}')

    def close_gw_port(self):
        """
                Close the connection to the Gateway.

                If the Gateway is connected, this function will close the port.
                """
        self.gateway_obj.close_port()
        self.gateway_obj.exit_gw_api()

    def start_test(self):
        """
        Initiates and executes the test sequence for Wiliot pixie.

        This function performs several key tasks:
        1. Initializes the data structures for storing test results in CSV format.
        2. Executes the test by calling the `run()` method on the `wiliot_tag_test` object.
        3. Collects statistics and summary information about the test.
        4. Populates the run data dictionary with test results and statistics.
        5. Updates the packets data CSV with packet-level information.
        6. If in 'online' mode, it also validates the test results by comparing them with cloud data.

        Returns:
        - True: If the test and optional validation are successful.
        - False: If the test or validation fails.

        Note: This function will also close the Gateway port in case of an exception.
        """
        try:
            self.initialize_csv_data()  # Init the run result dictionary
            res = self.wiliot_tag_test.run()  # Begin the run

            if res.is_results_empty():
                #   Failed
                self.logger.logger.error('Test failed')
                if self.tester_type == 'sample':
                    self.runDataDict['tested'] = 0
                else:
                    self.runDataDict['total_run_tested'] = 0
                self.runDataDict['responded'] = 0
                self.runDataDict['passed'] = 0
                raise Exception('something went wrong during test and test was not preformed')

            selected_tag_statistics = res.tests[0].selected_tag.get_statistics()  # Get all the statistics
            test_sum = res.tests[0].get_summary()

            if self.tester_type == 'sample':
                self.runDataDict['tested'] = 1
            else:
                self.runDataDict['total_run_tested'] = 1
            self.runDataDict['testTime'] = str(res.get_total_test_duration())
            self.runDataDict['passed'] = int(res.is_all_tests_passed())
            self.runDataDict['responded'] = int(1 if res.tests[0].all_packets.__len__() > 0 else 0)
            if self.tester_type == 'sample':
                self.runDataDict['responding[%]'] = str(
                    self.runDataDict['responded'] / self.runDataDict['tested'] * 100) + '%'
                self.runDataDict['passed[%]'] = str(self.runDataDict['passed'] / self.runDataDict['tested'] * 100) + '%'
            else:
                self.runDataDict['responding[%]'] = str(
                    self.runDataDict['responded'] / self.runDataDict['total_run_tested'] * 100) + '%'
                self.runDataDict['passed[%]'] = str(
                    self.runDataDict['passed'] / self.runDataDict['total_run_tested'] * 100) + '%'

            if self.runDataDict['passed']:
                self.logger.logger.info('Test passed')
            elif self.runDataDict['responded']:
                selected_tag_statistics = res.tests[0].all_packets.get_statistics()
                self.logger.logger.info('Test failed, tag responded but tbp could not be calculated')
            else:
                self.logger.logger.info('Test failed, tag did not respond')

            # save files:
            self.populate_run_data(selected_tag_statistics, test_sum)  # Fill run result dictionary
            self.update_packets_data(res, selected_tag_statistics)

            self.wiliot_tag_test.exit_tag_test()
            return True

        except Exception as e:
            self.close_gw_port()
            self.logger.logger.error(f'Problem starting the test {e}')
            return False

    def initialize_csv_data(self):
        """
                Initialize the CSV data structures.

                Sets up dictionaries for run data and packet data.
                """

        test_values = self.test_suite[self.test]['tests'][0]
        if self.tester_type == 'sample':
            self.runDataDict = {
                'testerStationName': 'WiliotSample',
                'commonRunName': self.common_run_name,
                'batchName': '',
                'testerType': self.tester_type,
                'comments': '',
                'errors': '',
                'timeProfile': test_values.get('timeProfile', ''),
                'txPower': test_values.get('absGwTxPowerIndex', ''),
                'energizingPattern': test_values.get('energizingPattern', ''),
                'tested': '0',
                'passed': '0',
                'yield': '0',
                'inlay': '',
                'responded': '0',
                'responding[%]': '0',
                'passed[%]': '0',
                'testStatus': 'False',
                'operator': 'Wiliot',
                'testTime': '',
                'runStartTime': '',
                'runEndTime': '',
                'antennaType': 'TIKI',
                'surface': '',
                'numChambers': '1',
                'gwVersion': self.gw_version,
                'pyWiliotVersion': self.pywiliot_version,
                'bleAttenuation': '',
                'loraAttenuation': '',
                'testTimeProfilePeriod': test_values.get('timeProfile'[1], ''),
                'testTimeProfileOnTime': test_values.get('timeProfile'[0], ''),
                'ttfpAvg': '',
                'tbpAvg': '',
                'tbpStd': '',
                'rssiAvg': '',
                'maxTtfp': '',
                'controlLimits': '',
                'hwVersion': '',
                'sub1gFrequency': '',
                'failBinStr': '',
                'failBin': ''
            }

            self.test_data = {
                'commonRunName': self.common_run_name,
                'encryptedPacket': '',
                'time': '',
                'reel': self.common_run_name,
                'ttfp': '',
                'tbp': '',
                'adv_address': '',
                'status': ''}
        else:
            self.runDataDict = {
                'testerStationName': 'WiliotSample',
                'commonRunName': self.common_run_name,
                'batchName': '',
                'testerType': self.tester_type,
                'comments': '',
                'errors': '',
                'timeProfile': test_values.get('timeProfile', ''),
                'txPower': test_values.get('absGwTxPowerIndex', ''),
                'energizingPattern': test_values.get('energizingPattern', ''),
                'total_run_tested': '0',
                'passed': '0',
                'yield': '0',
                'inlay': '',
                'responded': '0',
                'responding[%]': '0',
                'passed[%]': '0',
                'testStatus': 'False',
                'operator': 'Wiliot',
                'testTime': '',
                'runStartTime': '',
                'runEndTime': '',
                'antennaType': 'TIKI',
                'surface': '',
                'numChambers': '1',
                'gwVersion': self.gw_version,
                'pyWiliotVersion': self.pywiliot_version,
                'bleAttenuation': '',
                'loraAttenuation': '',
                'testTimeProfilePeriod': test_values.get('timeProfile'[1], ''),
                'testTimeProfileOnTime': test_values.get('timeProfile'[0], ''),
                'ttfpAvg': '',
                'test_suite_dict': self.test_suite[self.test],
                'owner_id': self.owner_id,
                'to_print': False,
                'qr_validation': False,
                'tbpAvg': '',
                'tbpStd': '',
                'rssiAvg': '',
                'maxTtfp': '',
                'controlLimits': '',
                'hwVersion': '',
                'sub1gFrequency': '',
                'failBinStr': '',
                'failBin': ''
            }

            self.test_data = {
                'common_run_name': self.common_run_name,
                'raw_packet': '',
                'adv_address': '',
                'time_from_start': '',
                'ttfp': '',
                'tbp_mean': '',
                'rssi': '',
                'gw_packet': '',
                'tag_run_location': '0',
                'external_id': '',
                'status_offline': '',
                'selected_tag': '',
                'fail_bin': '',
                'fail_bin_str': '',
                'test_num': 0,
                'is_test_pass': '',
                'packet_status': ''}

        dict_to_csv(dict_in=self.runDataDict, path=self.run_data_path, append=False, only_titles=True)
        dict_to_csv(dict_in=self.test_data, path=self.packets_data_path, append=False, only_titles=True)

    def populate_run_data(self, selected_tag_statistics, test_sum):
        """
                Populate the run data dictionary.

                Parameters:
                - selected_tag_statistics (dict): Statistics of the selected tag.
                - test_sum (dict): Summary of the test.

                Fills the run data dictionary with test results and statistics.
                """
        self.runDataDict['testStatus'] = bool(test_sum['is_test_pass'])
        self.runDataDict['runStartTime'] = test_sum['test_start_time'].strftime("%Y-%m-%d %H:%M:%S.%f")
        self.runDataDict['runEndTime'] = test_sum['test_end_time'].strftime("%Y-%m-%d %H:%M:%S.%f")
        self.runDataDict['ttfpAvg'] = str(selected_tag_statistics.get('ttfp', 'nan'))
        self.runDataDict['tbpAvg'] = str(selected_tag_statistics.get('tbp_mean', 'nan'))
        self.runDataDict['tbpStd'] = str(selected_tag_statistics.get('tbp_std', 'nan'))
        self.runDataDict['rssiAvg'] = str(selected_tag_statistics.get('rssi_mean', 'nan'))
        self.runDataDict['maxTtfp'] = str(selected_tag_statistics.get('ttfp', 'nan'))
        dict_to_csv(dict_in=self.runDataDict, path=self.run_data_path, append=True, only_titles=False)

    def update_packets_data(self, res=None, stats=None):
        """
                Update the packets data CSV.

                Parameters:
                - res (object): The result object containing test data.
                - stats (dict): Statistics of the selected tag.

                Updates the packets data CSV with packet information.
                """

        if res:
            r = res.tests[0]
            if r.selected_tag.__len__() > 0:  # pass
                tag_data = r.selected_tag
            elif r.all_packets.__len__() > 0:  # responded but did not pass
                tag_data = r.all_packets
            else:  # No response
                if self.tester_type == 'sample':
                    self.test_data.update({
                        'encryptedPacket': '',
                        'time': '',
                        'status': r.is_test_passed,
                        'ttfp': str(stats.get('ttfp', 'nan')),
                        'tbp': str(stats.get('tbp_min', 'nan')),
                        'adv_address': ''})
                else:
                    self.test_data.update({
                        'raw_packet': '',
                        'time_from_start': '',
                        'packet_status': r.is_test_passed,
                        'ttfp': str(stats.get('ttfp', 'nan')),
                        'tbp_mean': str(stats.get('tbp_min', 'nan')),
                        'rssi': str(stats.get('rssi', 'nan')),
                        'tag_run_location': '',
                        'status_offline': 0,
                        'gw_packet': '',
                        'fail_bin': FailureCodes.NO_RESPONSE.value,
                        'fail_bin_str': FailureCodes.NO_RESPONSE.name,
                        'is_test_pass': r.is_test_passed,
                        'adv_address': ''})
                return

            if self.tester_type == 'sample':
                for p in tag_data:
                    self.test_data.update({
                        'encryptedPacket': p.get_packet(),
                        'time': p.gw_data['time_from_start'][0] if p.gw_data['time_from_start'].size > 1 else p.gw_data[
                            'time_from_start'].item(),
                        'status': r.is_test_passed,
                        'ttfp': str(stats.get('ttfp', 'nan')),
                        'tbp': str(stats.get('tbp_min', 'nan')),
                        'adv_address': p.packet_data['adv_address']})
                    dict_to_csv(dict_in=self.test_data, path=self.packets_data_path, append=True, only_titles=False)
            else:
                for p in tag_data:
                    for i in range(p.gw_data['gw_packet'].size):
                        if p.gw_data["gw_packet"].size > 1:
                            gw_packet = p.gw_data["gw_packet"][i]
                            rssi = p.gw_data["rssi"][i]
                            time_from_start = p.gw_data["time_from_start"][i]
                        else:
                            gw_packet = p.gw_data["gw_packet"].item()
                            rssi = p.gw_data["rssi"].item()
                            time_from_start = p.gw_data["time_from_start"].item()
                        self.test_data.update({
                            'raw_packet': p.get_packet(),
                            'status_offline': 1,
                            'time_from_start': time_from_start,
                            'packet_status': r.is_test_passed,
                            'ttfp': str(stats.get('ttfp', 'nan')),
                            'tbp_mean': str(stats.get('tbp_min', 'nan')),
                            'rssi': rssi,
                            'gw_packet': gw_packet,
                            'fail_bin': FailureCodes.PASS.value,
                            'fail_bin_str': FailureCodes.PASS.name,
                            'is_test_pass': r.is_test_passed,
                            'selected_tag': p.packet_data['adv_address'],
                            'adv_address': p.packet_data['adv_address']})
                        dict_to_csv(dict_in=self.test_data, path=self.packets_data_path, append=True, only_titles=False)

    def cloud_upload(self):
        """
        Uploads the collected test data to the cloud for further analysis and storage.

        Parameters: - env (str): Specifies the cloud environment to upload to. It can be either 'prod' for production
        or 'test' for testing.

        The function performs the following tasks:
        1. Calls the `upload_to_cloud_api` function to upload both run data and packets data CSV files.
        2. Logs the status of the upload operation.

        Returns:
        - True: If the data upload to the cloud is successful.
        - False: If the data upload fails for any reason.

        Note: The function uses the owner ID and API key that were set during the cloud connection phase.
        """
        tester_type = self.tester_type + '-test'
        success = upload_to_cloud_api(batch_name='',
                                                  tester_type=tester_type,
                                                  run_data_csv_name=self.run_data_path,
                                                  packets_data_csv_name=self.packets_data_path,
                                                  is_path=True,
                                                  env=self.env,
                                                  owner_id=self.owner_id,
                                                  logger_=self.logger.logger.name)

        return success

    def get_common_run_name(self):
        return self.common_run_name

    def get_out_put_dir(self):
        return self.output


if __name__ == "__main__":
    Test = SimpleTest()



#  """
#    Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.
#
#    Redistribution and use of the Software in source and binary forms, with or without modification,
#     are permitted provided that the following conditions are met:
#
#       1. Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#
#       2. Redistributions in binary form, except as used in conjunction with
#       Wiliot's Pixel in a product or a Software update for such product, must reproduce
#       the above copyright notice, this list of conditions and the following disclaimer in
#       the documentation and/or other materials provided with the distribution.
#
#       3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
#       may be used to endorse or promote products or services derived from this Software,
#       without specific prior written permission.
#
#       4. This Software, with or without modification, must only be used in conjunction
#       with Wiliot's Pixel or with Wiliot's cloud service.
#
#       5. If any Software is provided in binary form under this license, you must not
#       do any of the following:
#       (a) modify, adapt, translate, or create a derivative work of the Software; or
#       (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
#       discover the source code or non-literal aspects (such as the underlying structure,
#       sequence, organization, ideas, or algorithms) of the Software.
#
#       6. If you create a derivative work and/or improvement of any Software, you hereby
#       irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
#       royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
#       right and license to reproduce, use, make, have made, import, distribute, sell,
#       offer for sale, create derivative works of, modify, translate, publicly perform
#       and display, and otherwise commercially exploit such derivative works and improvements
#       (as applicable) in conjunction with Wiliot's products and services.
#
#       7. You represent and warrant that you are not a resident of (and will not use the
#       Software in) a country that the U.S. government has embargoed for use of the Software,
#       nor are you named on the U.S. Treasury Department’s list of Specially Designated
#       Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
#       You must not transfer, export, re-export, import, re-import or divert the Software
#       in violation of any export or re-export control laws and regulations (such as the
#       United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
#       and use restrictions, all as then in effect
#
#     THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
#     OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
#     WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
#     QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
#     IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
#     ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
#     OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
#     FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
#     (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
#     (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
#     CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
#     (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
#     (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
#  """
import datetime
import os
import json
import threading
import multiprocessing
import time
import pandas as pd
import webbrowser
import requests
from queue import Queue

from wiliot_core import set_logger, GetApiKey, InlayTypes
from wiliot_tools.utils.wiliot_gui.wiliot_gui import *
from wiliot_tools.association_tool.association_configs import is_wiliot_code, is_asset_code
from wiliot_testers.association_tester.hw_components.scanner_component import Scanner
from wiliot_testers.association_tester.modules.performance_module import ReelVerification
from wiliot_testers.association_tester.modules.association_module import TagAssociation
from wiliot_testers.association_tester.modules.gui_module import AssociationAndVerificationGUI
pd.options.mode.chained_assignment = None  # default='warn'


GUI_USER_INPUT_PATH = os.path.dirname(os.path.abspath(__file__))
FILE_NAME = 'association_and_verification_gui_user_inputs.json'
GUI_FILE = os.path.abspath(os.path.join(GUI_USER_INPUT_PATH, 'configs', FILE_NAME))
RUN_PARAMS_FILE = os.path.abspath(os.path.join(GUI_USER_INPUT_PATH, 'configs', 'last_run_params.json'))
INLAY_TYPE = InlayTypes.TIKI_121.name
DEFAULT_VALUES = {
    'min_test_time': '1.5',
    'max_test_time': '5',
    'time_to_move': '0.5',
    'need_to_associate': 'no',
    'is_step_machine': 'yes',
    'asset_location': 'last',
    'owner_id': 'wiliot-ops',
    'category_id': '86fd9f07-38b0-466f-b717-2e285b803f5c',
    'energy_pattern': '51',
    'time_profile': '5,15',
    'ble_power': '22', 'sub1g_power': '29', 'sub1g_freq': '925000',
    'scan_ch': '37',
    'is_listen_bridge': 'yes',
    'env': 'prod',
    'scanner_type': 'Cognex',
    'sc_min_location': '30',
    'sc_association': '90',
    'sc_scanning': '90',
    'sc_responding': '90',
    'sc_n_no_scan': '2',
    'run_name': 'test'
}

DEFAULT_RUN_PARAM = {"last_run_name": "test", "last_location": 0, "last_asset_id": "", "last_asset_location": 0}

TITLE_FONT = ("Helvetica", 14, "bold")
SECTION_FONT = ("Helvetica", 12, "bold")


def get_params(file_path, default_values):
    if os.path.isfile(file_path):
        out = json.load(open(file_path, "rb"))
        for k in default_values.keys():
            if k not in out.keys():
                out[k] = default_values[k]
    else:
        out = default_values
    return out


def get_user_inputs():
    """
    opens GUI for selecting a file and returns it
    """
    # upload config
    default_values = get_params(GUI_FILE, DEFAULT_VALUES)
    run_params = get_params(RUN_PARAMS_FILE, DEFAULT_RUN_PARAM)

    params_dict = {

        # Setup Button
        'check_setup': {
            'text': 'Check Setup',
            'value': '',
            'widget_type': 'button',
            'group': 'Label Location'
        },

        # Label Location Section
        'run_name': {
            'text': 'Run name',
            'value': default_values['run_name'],
            'widget_type': 'entry',
            'group': 'Label Location'
        },
        'generate_run_name': {
            'text': 'Generate Run Name',
            'value': '',
            'widget_type': 'button',
            'group': 'Label Location'
        },
        'first_location': {
            'text': 'Start label location',
            'value': 0,
            'widget_type': 'entry',
            'group': 'Label Location'
        },
        'last_tested': {
            'value': f'last tested run name was: {run_params["last_run_name"]}\n'
                     f'last tested location was: {run_params["last_location"]}\n'
                     f'last scanned asset id was: {run_params["last_asset_id"]} '
                     f'at location: {run_params["last_asset_location"]}',
            'text': '',
            'widget_type': 'label',
            'group': 'Label Location'
        },

        # Tag-GW Configuration Section

        'min_test_time': {
            'text': 'Minimal wait time per location [sec]',
            'value': default_values['min_test_time'],
            'widget_type': 'entry',
            'group': 'Tag-GW Configuration'
        },
        'max_test_time': {
            'text': 'Maximal wait time per location [sec]',
            'value': default_values['max_test_time'],
            'widget_type': 'entry',
            'group': 'Tag-GW Configuration'
        },
        'energy_pattern': {
            'text': 'Energy Pattern',
            'value': default_values['energy_pattern'],
            'widget_type': 'entry',
            'group': 'Tag-GW Configuration'
        },
        'time_profile': {
            'text': 'Time Profile',
            'value': default_values['time_profile'],
            'widget_type': 'entry',
            'group': 'Tag-GW Configuration'
        },
        'ble_power': {
            'text': 'BLE Power[dBm]',
            'value': default_values['ble_power'],
            'widget_type': 'entry',
            'group': 'Tag-GW Configuration'
        },
        'sub1g_power': {
            'text': 'Sub1G Power[dBm]',
            'value': default_values['sub1g_power'],
            'widget_type': 'entry',
            'group': 'Tag-GW Configuration'
        },
        'sub1g_freq': {
            'text': 'Sub1G frequency[kHz]',
            'value': default_values['sub1g_freq'],
            'widget_type': 'entry',
            'group': 'Tag-GW Configuration'
        },
        'is_listen_bridge': {
            'text': 'listen to Bridge?',
            'value': default_values['is_listen_bridge'],
            'options': ('yes', 'no'),
            'widget_type': 'combobox',
            'group': 'Tag-GW Configuration'
        },

        # Scanner Section
        'scanner_type': {
            'text': 'Scanner Type',
            'value': default_values['scanner_type'],
            'options': ('Cognex', ''),
            'widget_type': 'combobox',
            'group': 'Scanner'
        },
        'asset_location': {
            'text': 'Asset location with respect to Wiliot code',
            'value': default_values['asset_location'],
            'options': ('first', 'last'),
            'widget_type': 'combobox',
            'group': 'Scanner'
        },

        # Reel-to-Reel Section
        'is_step_machine': {
            'text': 'Is step machine?',
            'value': default_values['is_step_machine'],
            'options': ('yes', 'no'),
            'widget_type': 'combobox',
            'group': 'Reel-to-Reel'
        },
        'time_to_move': {
            'text': 'Movement time [sec]',
            'value': default_values['time_to_move'],
            'widget_type': 'entry',
            'group': 'Reel-to-Reel'
        },

        # Association Section
        'need_to_associate': {
            'text': 'Do association?',
            'value': default_values['need_to_associate'],
            'options': ('yes', 'no'),
            'widget_type': 'combobox',
            'group': 'Association'
        },
        'owner_id': {
            'text': 'Owner id',
            'value': default_values['owner_id'],
            'widget_type': 'entry',
            'group': 'Association'
        },
        'env': {
            'text': 'Environment',
            'value': default_values['env'],
            'widget_type': 'entry',
            'group': 'Association'
        },
        'category_id': {
            'text': 'Asset category id',
            'value': default_values['category_id'],
            'widget_type': 'entry',
            'group': 'Association'
        },

        # Stop Criteria Section
        'sc_min_location': {
            'text': 'Minimum tags to test before applying "yield-stop-criteria"',
            'value': default_values['sc_min_location'],
            'widget_type': 'entry',
            'group': 'Stop Criteria'
        },
        'sc_association': {
            'text': 'Stop if association yield is lower than [%]',
            'value': default_values['sc_association'],
            'widget_type': 'entry',
            'group': 'Stop Criteria'
        },
        'sc_scanning': {
            'text': 'Stop if scanning yield is lower than [%]',
            'value': default_values['sc_scanning'],
            'widget_type': 'entry',
            'group': 'Stop Criteria'
        },
        'sc_responding': {
            'text': 'Stop if responding yield is lower than [%]',
            'value': default_values['sc_responding'],
            'widget_type': 'entry',
            'group': 'Stop Criteria'
        },
        'sc_n_no_scan': {
            'text': 'Stop if N successive labels are failed to be scanned',
            'value': default_values['sc_n_no_scan'],
            'widget_type': 'entry',
            'group': 'Stop Criteria'
        },

        # Run Button
        'run_button': {
            'text': 'Run',
            'value': '',
            'widget_type': 'button',
            'columnspan': 2

        },
    }
    scanner = None
    is_setup_valid = False
    scanned_codes = ['test']

    def on_check_setup():
        values = gui.get_all_values()
        nonlocal scanner
        is_setup_valid, scanner, scanned_codes = is_valid_setup(values, scanner, gui)

    def on_generate_run_name():
        scanned_codes.sort(reverse=True)
        gui.update_widget(widget_key='run_name', new_value='_'.join(scanned_codes))

    def on_run():
        values = gui.get_all_values()
        nonlocal scanner
        if not is_setup_valid:
            is_valid, scanner, scanned_codes = is_valid_setup(values, scanner, gui)
            if not is_valid:
                pass
        for k, v in values.items():
            default_values[k] = v
        with open(GUI_FILE, 'w') as f:
            json.dump(default_values, f)
        with open(RUN_PARAMS_FILE, 'w') as f:
            json.dump(run_params, f)

        if scanner is not None:
            scanner.disconnect()
        return default_values

    gui = WiliotGui(params_dict, do_button_config=False, height_offset=-50)
    gui.widgets['last_tested'].configure(anchor="w", justify='left')
    gui.add_event(widget_key='generate_run_name', command=on_generate_run_name, event_type='button')
    gui.add_event(widget_key='check_setup', command=on_check_setup, event_type='button')
    gui.add_event(widget_key='run_button', command=on_run, event_type='button')
    gui.run()


def is_valid_setup(values, scanner, gui):
    # check scanner
    is_valid = 'yes'
    scanned_codes = ['test']
    if values['scanner_type'].lower() == 'cognex':
        if scanner is None:
            scanner = Scanner()
        scanned_codes = scanner.scan()
        yes_or_no_layout = {
            'question': {
                'value': f'Starting location was set to: {values["first_location"]}\n'
                         f'The scanned codes are:\n{scanned_codes}\n'
                         f'Is the starting location is correct?\n'
                         f'Are those the codes of the first label?\n\n'
                         f'If not, please click on No, try to re-position scanner and try again',
                'text': '',
                'widget_type': 'label',
            },
            'yes_button': {
                'text': 'Yes',
                'value': '',
                'widget_type': 'button',
            },
            'no_button': {
                'text': 'No',
                'value': '',
                'widget_type': 'button',
            },
        }

        def on_no_button():
            print('re-positioning the scanner and try again')
            yes_or_no_gui.layout.destroy()
            return False, scanner, scanned_codes

        def on_yes_button():
            client_types = ['asset', None]
            for client_type in client_types:
                g = GetApiKey(gui_type='ttk',
                              env=values['env'],
                              owner_id=values['owner_id'],
                              client_type=client_type
                              )
                api_key = g.get_api_key()
                if not api_key:
                    file_path = g.get_config_path()
                    popup_message(
                        f'Could not found an api key for owner id {values["owner_id"]} and env {values["env"]}'
                        f'at path: {file_path}')
                    yes_or_no_gui.layout.destroy()
                    return False, scanner, scanned_codes
            yes_or_no_gui.layout.destroy()
            return True, scanner, scanned_codes

        yes_or_no_gui = WiliotGui(params_dict=yes_or_no_layout, parent=gui.layout, do_button_config=False)
        yes_or_no_gui.add_event(widget_key='no_button', command=on_no_button, event_type='button')
        yes_or_no_gui.add_event(widget_key='yes_button', command=on_yes_button, event_type='button')
        yes_or_no_gui.run()


class AssociationAndVerificationTester(object):
    def __init__(self, user_input):
        start_time = datetime.datetime.now()
        common_run_name = f"{user_input['run_name']}_{start_time.strftime('%Y%m%d_%H%M%S')}"
        self.logger_path, self.logger = set_logger(app_name='AssociationAndVerificationTester',
                                                   common_run_name=common_run_name)
        self.run_data = {'common_run_name': common_run_name,
                         'tester_type': 'association_and_verification',
                         'inlay_type': INLAY_TYPE,
                         'station_name': os.environ['testerStationName']
                         if 'testerStationName' in os.environ else 'testStation',
                         'run_start_time': start_time}
        stop_event = multiprocessing.Event()
        exception_q = Queue(maxsize=100)
        self.app_running = True
        self.performance = ReelVerification(user_input=user_inputs,
                                            stop_event=stop_event,
                                            is_app_running=self.is_app_running,
                                            logger_name=self.logger.name,
                                            logger_path=self.logger_path,
                                            exception_q=exception_q)
        self.association = TagAssociation(user_input=user_inputs,
                                          stop_event=stop_event,
                                          is_app_running=self.is_app_running,
                                          logger_name=self.logger.name,
                                          logger_path=self.logger_path,
                                          exception_q=exception_q,
                                          run_param_file=RUN_PARAMS_FILE)

        self.gui = AssociationAndVerificationGUI(logger_name=self.logger.name,
                                                 get_data_func=self.get_data,
                                                 get_stat_func=self.get_stat,
                                                 common_run_name=common_run_name)
        self.user_inputs = user_input
        self.stop_event = stop_event
        self.exception_q = exception_q
        self.results_df = pd.DataFrame()
        self.current_stat = {'n_locations': '0',
                             'scan_success': '0%', 'association_success': '0%', 'responding_rate': '0%',
                             'n_success': '0', 'success_rate': '0%'}
        self.duplicated_code = pd.DataFrame()
        self.neglected_duplication = pd.DataFrame()
        self.performance_thread = None
        self.association_thread = None
        self.gui_thread = None

    def run(self):
        self.performance_thread = threading.Thread(target=self.performance.run_app, args=())
        self.association_thread = threading.Thread(target=self.association.run_app, args=())
        self.gui_thread = threading.Thread(target=self.gui.run_app, args=())
        self.performance_thread.start()
        self.association_thread.start()
        self.gui_thread.start()
        webbrowser.open(self.gui.get_url())

        self.run_app()

    def run_app(self):
        while True:
            try:
                time.sleep(1)
                if self.gui.is_stopped_by_user():
                    self.logger.info('run stopped by user')
                    break
                elif self.app_running != self.gui.is_app_running():
                    self.logger.info(f'run was {"paused" if self.app_running else "continued"} by user')
                    self.app_running = not self.app_running

                if not self.exception_q.empty():
                    self.app_running = False
                    self.logger.warning('got exception during run, pause the app')
                    self.handle_exceptions()

                if self.app_running:
                    df_ass = self.association.get_locations_df()
                    df_per = self.performance.get_packets_df()
                    self.merge_results(location_df=df_ass, packets_df=df_per)
                    self.current_stat = self.calc_stat()
                    self.handle_stop_criteria()
                    # save data:
                    self.save_location_data()
                    self.save_run_data()
                else:
                    time.sleep(1)
            except Exception as e:
                self.logger.warning(f'got exception during run: {e}')
                if not self.exception_q.full():
                    self.exception_q.put(f'Main: {e}')

        # stop run
        self.stop()

    def is_app_running(self):
        return self.app_running

    def handle_stop_criteria(self):
        if int(self.current_stat['n_successive_bad_scan']) >= int(self.user_inputs['sc_n_no_scan']):
            raise Exception(f"{self.current_stat['n_successive_bad_scan']} successive bad scanning were detected. "
                            f"Check if machine is stuck")

        if self.current_stat['is_duplicated']:
            self.neglected_duplication = self.duplicated_code.copy()
            raise Exception(f"{self.duplicated_code.values} codes duplications were detected. "
                            f"Check if machine is stuck")

        if int(self.current_stat['n_locations']) < int(self.user_inputs['sc_min_location']):
            return
        if self.user_inputs['need_to_associate'].lower() == 'yes':
            if float(self.current_stat['association_success'].replace('%', '')) < float(
                    self.user_inputs['sc_association']):
                self.user_inputs['sc_association'] = 0  # pop only once
                raise Exception(f"association yield of {self.current_stat['association_success']} was detected")
        if float(self.current_stat['scan_success'].replace('%', '')) < float(self.user_inputs['sc_scanning']):
            self.user_inputs['sc_scanning'] = 0  # pop only once
            raise Exception(f"scan yield of {self.current_stat['scan_success']} was detected")
        if float(self.current_stat['responding_rate'].replace('%', '')) < float(self.user_inputs['sc_responding']):
            self.user_inputs['sc_responding'] = 0  # pop only once
            raise Exception(f"responding yield of {self.current_stat['responding_rate']} was detected")

    def handle_exceptions(self):
        n_exceptions = self.exception_q.qsize()
        exceptions_str = []
        for _ in range(n_exceptions):
            exceptions_str.append(self.exception_q.get())
        self.logger.warning('\n'.join(exceptions_str))
        popup_message('\n'.join(exceptions_str))

    def print_run(self):  # for debug
        t_i = time.time()
        while time.time() - t_i < 60:
            try:
                time.sleep(2)
                df_ass = self.association.get_locations_df()
                print(f'n unique locations: {len(df_ass)}')
                df_per = self.performance.get_packets_df()
                print(f'n unique adva: {len(df_per)}')
                self.merge_results(location_df=df_ass, packets_df=df_per)
                print(f'n all results: {len(self.results_df)}')
                stat = self.calc_stat()
                print(stat)
            except Exception as e:
                self.logger.warning(f'got exception during print_run: {e}')

        # stop run
        self.stop_event.set()
        self.stop()

    def calc_stat(self):
        stat_out = {}
        rel_data = self.results_df.loc[~(self.results_df['location'].isna()) & ~(self.results_df['location'] == '')]
        n_location = len(rel_data)
        stat_out['n_locations'] = str(n_location)
        stat_out['n_tags_outside_test'] = str(len(self.results_df) - n_location)
        if n_location > 0:
            stat_out['scan_success'] = f'{round(rel_data["scan_status"].sum() / n_location * 100, 2)}%'
            ass_valid = rel_data["is_associated"][rel_data["associate_status_code"] != '']
            stat_out['association_success'] = f'{round(ass_valid.sum() / max([ass_valid.count(), 1]) * 100, 2)}%'
            if 'n_packets' in rel_data.keys():
                respond_valid = rel_data["n_packets"][rel_data["wiliot_code"] != '']
                stat_out['responding_rate'] = f'{round(respond_valid.notna().sum() / len(respond_valid) * 100, 2)}%'
            else:
                stat_out['responding_rate'] = '0%'
            stat_out['n_successive_bad_scan'] = \
                rel_data['location'].iloc[-1] - rel_data['location'][rel_data['scan_status']].iloc[-1] \
                    if any(rel_data['scan_status']) else n_location

            df_per_loc = rel_data.drop_duplicates(subset=['location'])
            self.duplicated_code = pd.concat([
                df_per_loc['asset_code'].loc[
                    (df_per_loc.duplicated(subset=['asset_code']) & (df_per_loc['asset_code'] != ''))],
                df_per_loc['wiliot_code'].loc[
                    (df_per_loc.duplicated(subset=['wiliot_code'])) & (df_per_loc['wiliot_code'] != '')]
            ], axis=0)
            stat_out['is_duplicated'] = len(self.duplicated_code) > len(self.neglected_duplication)
            if 'is_success' in rel_data.keys():
                stat_out['n_success'] = f'{rel_data["is_success"].sum()}'
                stat_out['success_rate'] = f'{round(rel_data["is_success"].sum() / n_location * 100, 2)}%'
            else:
                stat_out['n_success'] = '0'
                stat_out['success_rate'] = '0%'
        else:
            stat_out['scan_success'] = '0%'
            stat_out['association_success'] = '0%'
            stat_out['responding_rate'] = '0%'
            stat_out['n_successive_bad_scan'] = 0
            stat_out['is_duplicated'] = False
            stat_out['n_success'] = '0'
            stat_out['success_rate'] = '0%'

        return stat_out

    def shutdown_server(self):
        try:
            requests.post(f'{self.gui.get_url()}shutdown')
        except Exception as e:
            pass

    def stop(self):
        self.stop_event.set()
        self.shutdown_server()

        self.performance_thread.join(15)
        if self.performance_thread.is_alive():
            self.logger.warning('performance thread is still running')
        self.association_thread.join(15)
        if self.association_thread.is_alive():
            self.logger.warning('association thread is still running')
        self.gui_thread.join(15)
        if self.gui_thread.is_alive():
            self.logger.warning('GUI thread is still running')

        # summary
        df_ass = self.association.get_locations_df()
        df_per = self.performance.get_packets_df()
        self.merge_results(location_df=df_ass, packets_df=df_per)
        self.current_stat = self.calc_stat()

        self.save_location_data()
        self.save_run_data()

        # handle exceptions:
        if not self.exception_q.empty():
            self.handle_exceptions()

        # show results:
        stat_out_str = "\n".join([f'{k}: {v}' for k, v in self.current_stat.items()])
        popup_message(f'{self.run_data["common_run_name"]}\n\n{stat_out_str}')

    def save_run_data(self):
        df_path = os.path.join(os.path.dirname(self.logger_path), f'{self.run_data["common_run_name"]}@run_data.csv')
        print(f'saving run config at: {df_path}')
        data_to_save = pd.DataFrame({**self.run_data, **{'end_run_time': datetime.datetime.now()}, **self.current_stat,
                                     **self.user_inputs}, index=[0])
        data_to_save.to_csv(df_path, index=False)

    def save_location_data(self):
        df_path = os.path.join(os.path.dirname(self.logger_path),
                               f'{self.run_data["common_run_name"]}@packets_data.csv')
        print(f'saving data at: {df_path}')
        data_to_save = self.results_df.loc[~self.results_df['location'].isna()]
        data_to_save.sort_values(by='location', inplace=True)
        data_to_save.to_csv(df_path, index=False)

    def get_data(self):
        return self.results_df

    def get_stat(self):
        return self.current_stat

    def merge_results(self, location_df, packets_df):
        if packets_df.empty and location_df.empty:
            return

        merged_df = None
        if packets_df.empty:
            merged_df = location_df
        else:
            packets_df.index = packets_df.index.set_names(['adv_address'])
            packets_df = packets_df.reset_index()

        if location_df.empty:
            merged_df = packets_df
            merged_df.insert(loc=0, column='location', value='')
            if 'n_packets' not in merged_df.keys():
                merged_df.insert(loc=0, column='n_packets', value=0)

        if merged_df is None:
            merged_df = pd.merge(location_df, packets_df, left_on='wiliot_code', right_on='external_id', how='outer')
            if 'n_packets' in merged_df.keys():
                is_responded = merged_df['n_packets'].apply(lambda x: int(x) > 0 if not pd.isnull(x) else False)
            else:
                is_responded = pd.Series([False] * merged_df.shape[0])
            if self.user_inputs['need_to_associate'].lower() == 'yes':
                is_success = pd.DataFrame([merged_df['scan_status'], merged_df['is_associated'], is_responded]).all()
            else:
                is_success = pd.DataFrame([merged_df['scan_status'], is_responded]).all()
            merged_df.insert(loc=len(merged_df.columns), column='is_success', value=is_success)
            merged_df.insert(loc=0, column='common_run_name', value=self.run_data['common_run_name'])

        self.results_df = merged_df


if __name__ == '__main__':
    user_inputs = get_user_inputs()
    print(user_inputs)

    av = AssociationAndVerificationTester(user_input=user_inputs)
    av.run()

    print('done')



"""
  Copyright (c) 2016- 2024, Wiliot Ltd. All rights reserved.

  Redistribution and use of the Software in source and binary forms, with or without modification,
   are permitted provided that the following conditions are met:

     1. Redistributions of source code must retain the above copyright notice,
     this list of conditions and the following disclaimer.

     2. Redistributions in binary form, except as used in conjunction with
     Wiliot's Pixel in a product or a Software update for such product, must reproduce
     the above copyright notice, this list of conditions and the following disclaimer in
     the documentation and/or other materials provided with the distribution.

     3. Neither the name nor logo of Wiliot, nor the names of the Software's contributors,
     may be used to endorse or promote products or services derived from this Software,
     without specific prior written permission.

     4. This Software, with or without modification, must only be used in conjunction
     with Wiliot's Pixel or with Wiliot's cloud service.

     5. If any Software is provided in binary form under this license, you must not
     do any of the following:
     (a) modify, adapt, translate, or create a derivative work of the Software; or
     (b) reverse engineer, decompile, disassemble, decrypt, or otherwise attempt to
     discover the source code or non-literal aspects (such as the underlying structure,
     sequence, organization, ideas, or algorithms) of the Software.

     6. If you create a derivative work and/or improvement of any Software, you hereby
     irrevocably grant each of Wiliot and its corporate affiliates a worldwide, non-exclusive,
     royalty-free, fully paid-up, perpetual, irrevocable, assignable, sublicensable
     right and license to reproduce, use, make, have made, import, distribute, sell,
     offer for sale, create derivative works of, modify, translate, publicly perform
     and display, and otherwise commercially exploit such derivative works and improvements
     (as applicable) in conjunction with Wiliot's products and services.

     7. You represent and warrant that you are not a resident of (and will not use the
     Software in) a country that the U.S. government has embargoed for use of the Software,
     nor are you named on the U.S. Treasury Department’s list of Specially Designated
     Nationals or any other applicable trade sanctioning regulations of any jurisdiction.
     You must not transfer, export, re-export, import, re-import or divert the Software
     in violation of any export or re-export control laws and regulations (such as the
     United States' ITAR, EAR, and OFAC regulations), as well as any applicable import
     and use restrictions, all as then in effect

   THIS SOFTWARE IS PROVIDED BY WILIOT "AS IS" AND "AS AVAILABLE", AND ANY EXPRESS
   OR IMPLIED WARRANTIES OR CONDITIONS, INCLUDING, BUT NOT LIMITED TO, ANY IMPLIED
   WARRANTIES OR CONDITIONS OF MERCHANTABILITY, SATISFACTORY QUALITY, NONINFRINGEMENT,
   QUIET POSSESSION, FITNESS FOR A PARTICULAR PURPOSE, AND TITLE, ARE DISCLAIMED.
   IN NO EVENT SHALL WILIOT, ANY OF ITS CORPORATE AFFILIATES OR LICENSORS, AND/OR
   ANY CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
   OR CONSEQUENTIAL DAMAGES, FOR THE COST OF PROCURING SUBSTITUTE GOODS OR SERVICES,
   FOR ANY LOSS OF USE OR DATA OR BUSINESS INTERRUPTION, AND/OR FOR ANY ECONOMIC LOSS
   (SUCH AS LOST PROFITS, REVENUE, ANTICIPATED SAVINGS). THE FOREGOING SHALL APPLY:
   (A) HOWEVER CAUSED AND REGARDLESS OF THE THEORY OR BASIS LIABILITY, WHETHER IN
   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE);
   (B) EVEN IF ANYONE IS ADVISED OF THE POSSIBILITY OF ANY DAMAGES, LOSSES, OR COSTS; AND
   (C) EVEN IF ANY REMEDY FAILS OF ITS ESSENTIAL PURPOSE.
"""
from time import sleep
from collections import Counter
import json
import pandas as pd
import numpy as np
from pathlib import Path

from wiliot_core import WiliotGateway, ActionType, DataType
from wiliot_core import PacketList

from wiliot_tools.test_equipment.test_equipment import Attenuator, EquipmentError
from wiliot_testers.tester_utils import TagsHandling, update_json_field

simulation = False
fine_only = False


def set_attn_power(attn_obj, attn_power, simulation=simulation):
    """
    configure Attenuator to a specific value
    gets:
        attn_obj:       Attenuator obj
        attn_power:     value to set attn

    return:
        status if Attenuator is set correctly

    """
    if simulation:
        status = True
    else:
        status = True
        print('Setting Attenuator to {attn_power}dB'.format(attn_power=attn_power))
        attn_obj.Setattn(attn_power)
        sleep(2)
        
        attn_current_config = attn_obj.Getattn()
        if (float(attn_current_config) - attn_power) != 0:
            print('Error setting ATTENUATOR')
            status = False
        print(
            "Attenuator is set to: {attn_current_config} dB".format(
                attn_current_config=attn_current_config.split('.')[0]))
    return status


def start_harvesting(gw_obj, num_of_packets=100, loops_limit=int(100 * 1.5),
                     energy_pattern_val=20):
    """
    configure GW and start rx proccess, gets a specific number of packets
    gets:
        gw_obj:             Gateway obj
        num_of_packets:     number of packets to be collected using GW

    return:
        Dataframe containing packets and data
    """
    if num_of_packets != 100:
        loops_limit = int(num_of_packets * 1.5)
    print('extract {num_of_packets} packets'.format(num_of_packets=num_of_packets))
    missing_packets = num_of_packets
    num_of_loops = loops_limit
    
    gw_obj.config_gw(pacer_val=0, energy_pattern_val=energy_pattern_val, time_profile_val=[5, 15],
                     received_channel=37, beacons_backoff_val=0)
    
    output_type = DataType.PACKET_LIST
    
    packet_data = PacketList()
    gw_obj.reset_listener()
    
    while missing_packets > 0 and num_of_loops > 0:
        num_of_loops -= 1
        sleep(1)
        if gw_obj.is_data_available():
            
            if num_of_loops == 0:
                gw_answer = gw_obj.get_packets(action_type=ActionType.ALL_SAMPLE,
                                               data_type=output_type)
            else:
                gw_answer = gw_obj.get_packets(action_type=ActionType.FIRST_SAMPLES, num_of_packets=num_of_packets,
                                               data_type=output_type, is_blocking=False)
                # gw_answer = gw_obj.get_packets(action_type=ActionType.FIRST_SAMPLES, num_of_packets=missing_packets,
                #                                data_type=output_type)
            
            missing_packets = num_of_packets - gw_answer.size()
            
            packet_data = packet_data + gw_answer
    
    # gw_obj.stop_continuous_listener()
    
    gw_obj.config_gw(time_profile_val=[0, 6])
    
    # df = pd.DataFrame(packet_data)
    total_loops = loops_limit - num_of_loops
    return packet_data, total_loops


def packet_data_analysis(packet_data):
    """
    Analyze packet data and get primary data
    gets:
        packet_data:    Dataframe containing packet data

    return:  tuple with 2 cells:
        primary tag adv
        percentage of primary tag packets
        secondary tag adv
        percentage of secondary tag packets
        rssi of primary tag packets
        rssi of secondary tag packets

    """
    packets_df = packet_data.get_df()
    tag_ids = packets_df['adv_address']
    tag_ids_repetitions = dict(Counter(tag_ids))
    
    total_packets = sum(tag_ids_repetitions.values())
    
    primary_tag = max(tag_ids_repetitions, key=tag_ids_repetitions.get)
    primary_tag_ratio = round(tag_ids_repetitions[primary_tag] / total_packets, 2)
    secondary_tag_ids_repetitions = tag_ids_repetitions.copy()
    
    secondary_tag = None
    secondary_tag_ratio = 0
    rssi = None
    
    if len(secondary_tag_ids_repetitions) > 1:
        secondary_tag_ids_repetitions.pop(primary_tag)
        secondary_tag = max(secondary_tag_ids_repetitions, key=secondary_tag_ids_repetitions.get)
        secondary_tag_ratio = round(tag_ids_repetitions[secondary_tag] / total_packets, 2)
    
    print("primary_tag", primary_tag)
    print("secondary_tag", secondary_tag)
    
    primary_tag_rows = packets_df[packets_df['adv_address'] == primary_tag]
    rssi = 0
    if len(primary_tag_rows) > 0 and primary_tag != '':
        rssi = None if np.isnan(np.mean(primary_tag_rows['rssi'])) else round(float(np.mean(primary_tag_rows['rssi'])),
                                                                              2)
    secondary_rssi = 0
    if secondary_tag != '':
        secondary_tag_rows = packets_df[packets_df['adv_address'] == secondary_tag]
        if len(secondary_tag_rows) > 0:
            secondary_rssi = None if np.isnan(np.mean(secondary_tag_rows['rssi'])) else round(float(
                np.mean(secondary_tag_rows['rssi'])), 2)
    
    return primary_tag, primary_tag_ratio, secondary_tag, secondary_tag_ratio, rssi, secondary_rssi


def sprinkler_harvesting(gw_obj, loops_limit=10, energy_pattern_val=20):
    """
    Setting GW to get duplicate packets, looking for 3 duplicates from the same tag,
        calculate minimum delta between packets
    gets:
        gw_obj:         Gateway object
        tag_adv:        tag advertising address to filter packets by
        loops_limit:    limit time to get packets data

    return:
        duplicates_delta_min:    time in msec
    """
    gw_obj.config_gw(pacer_val=0, energy_pattern_val=energy_pattern_val, time_profile_val=[5, 15],
                     received_channel=37, beacons_backoff_val=0)  # filter_val=True allow to get duplicate packets
    
    gw_obj.reset_listener()
    
    packet_data = PacketList()
    
    search_duplicates = True
    
    while search_duplicates and loops_limit > 0:
        sleep(5)
        
        gw_answer = gw_obj.get_packets(action_type=ActionType.ALL_SAMPLE, data_type=DataType.PACKET_LIST)
        
        packet_data = packet_data + gw_answer
        
        if any([len(p) >= 3 for p in packet_data.packet_list]):
            search_duplicates = False
        else:
            loops_limit = loops_limit - 1
    
    # gw_obj.stop_continuous_listener()
    # print('stopped processes')
    # gw_obj.reset_buffer()
    # gw_obj.stop_continuous_listener()
    gw_obj.config_gw(time_profile_val=[0, 6])
    
    return packet_data


def sweep_range(attn_values, attn_obj, gw_obj, tag_handle, energy_pattern_val=20, num_of_packets=100):
    """
    Loop over attenuator values and gather packet_data_analysis and sprinkler_analysis functions data
    :type attn_values: range
    :param attn_values: List of attenuation to sweep by
    :type attn_obj: Attenuator object
    :param attn_obj: Attenuator object
    :type gw_obj: Gateway object
    :param gw_obj: Gateway object
    :type tag_handle: TagsHandling object (wiliot_testers.testers_utils)
    :param tag_handle: TagsHandling object
    :type energy_pattern_val: int
    :param energy_pattern_val: energy pattern to check
    :type num_of_packets: int
    :param num_of_packets: how many packets to count


    return:
        attn_dict:
            Dictionary,
                keys: attn_value
                values: List including the following:
                    number of packets
                    percentage of primary tag packets
                    percentage of secondary tag packets
                    rssi of primary tag packets
                    rssi of secondary tag packets
                    seconds to get @num_of_packets packets
                    primary sprinkler min time
                    secondary sprinkler min time

    """
    attn_dict = {}
    primary_tag_adv = None
    secondary_tag_adv = None
    primary_tag_dist = None
    secondary_tag_dist = None
    for attn_power in attn_values:
        tag_handle.set_new_location()
        set_attn_status = set_attn_power(attn_obj, attn_power)
        if not simulation:
            print('ATTN is set to ', attn_obj.Getattn())
            if set_attn_status is not True:
                continue
        
        packet_data, total_loops = start_harvesting(gw_obj=gw_obj, num_of_packets=num_of_packets,
                                                    energy_pattern_val=energy_pattern_val)
        if packet_data.size() == 0:
            attn_dict[attn_power] = (0, 0, 0, 0, 0, 0, 0)
        else:
            (primary_tag_adv, primary_tag_dist, secondary_tag_adv, secondary_tag_dist, primary_avg_rssi,
             secondary_avg_rssi) = packet_data_analysis(packet_data)
        
        sprinkler_packet_data = sprinkler_harvesting(gw_obj, loops_limit=20, energy_pattern_val=energy_pattern_val)
        if len(sprinkler_packet_data) > 0:
            sprinkler_packet_primary = sprinkler_packet_data.filter_packet_by(values=primary_tag_adv)
            primary_avg_rssi = sprinkler_packet_primary.get_avg_rssi()
            primary_sprinkler_avg_tbp = sprinkler_packet_primary.get_avg_tbp(ignore_outliers=True)
            
            sprinkler_packet_secondary = sprinkler_packet_data.filter_packet_by(values=secondary_tag_adv)
            secondary_avg_rssi = sprinkler_packet_secondary.get_avg_rssi()
            secondary_sprinkler_avg_tbp = sprinkler_packet_secondary.get_avg_tbp(ignore_outliers=True)
            
            attn_dict[attn_power] = [len(packet_data),
                                     primary_tag_dist, secondary_tag_dist, primary_avg_rssi, secondary_avg_rssi,
                                     total_loops, primary_sprinkler_avg_tbp, secondary_sprinkler_avg_tbp]
            
            print(attn_dict)
            print('Waiting for brownout')
            sleep(5)
        else:
            attn_dict[attn_power] = (0, 0, 0, 0, 0, 0, 0)
    return attn_dict


def start_calibration(inlay_type='Single Band', num_of_packets=100, td_lower_limit=0, set_optimal_attn=True):
    """
    calibration process
    :type inlay_type: string
    :param inlay_type: will determine the energizing patterns we will use
    :type num_of_packets: int
    :param num_of_packets: default number of packets to analyze by
    :type td_lower_limit: double
    :param td_lower_limit: set sprinkler time delta to filter results by
    :type set_optimal_attn: Bool
    :param set_optimal_attn: A flag to decide if set attn to optimal value

    return:
        None
    """
    if inlay_type == 'Single Band':
        energy_pattern_values = [18]
    else:
        energy_pattern_values = [18, 51]
    
    attn_obj = None
    # create equipment objects
    if not simulation:
        try:
            attn_obj = Attenuator('API').GetActiveTE()
            current_attn = attn_obj.Getattn()
        except Exception as e:
            raise EquipmentError('Attenuator Error - Verify Attenuator connection')
    
    gw_obj = None
    try:
        gw_obj = WiliotGateway(auto_connect=True, logger_name='root', verbose=False)
        # gw_obj.openPort(port='COM4', baud=921600)
        gw_obj.write('!set_tester_mode 1')
        gw_obj.start_continuous_listener()
        if not gw_obj.get_connection_status()[0]:
            raise EquipmentError('Gateway Error - Verify WiliotGateway connection')
    except Exception as e:
        if gw_obj is not None:
            gw_obj.close_port()
        raise EquipmentError('Gateway Error - Verify WiliotGateway connection')
    
    tag_handle = TagsHandling(tags_list_len=np.inf, rssi_threshold=100, add_to_black_list_after_locations=np.inf)
    fine_attn_dict = {}
    for energy_pattern in energy_pattern_values:
        if fine_only:
            attn_values = range(15)
            coarse_attn_dict = {}
        else:
            print('Start sweep on coarse range:')
            attn_values = range(0, 11, 3)
            coarse_attn_dict = sweep_range(attn_values, attn_obj, gw_obj, tag_handle, energy_pattern_val=energy_pattern,
                                           num_of_packets=num_of_packets)
            
            print('Coarse data: {attn_dict}'.format(attn_dict=coarse_attn_dict))
            
            attn_dist = [value[0] for key, value in coarse_attn_dict.items()]
            
            top_values_keys = [list(coarse_attn_dict.keys())[i] for i in np.argsort(attn_dist)[-4:]]
            top_values_keys.remove(max(top_values_keys))
            
            attn_values = range(min(top_values_keys), max(top_values_keys) + 1, 1)
        
        print('Start sweep on fine range:')
        fine_attn_dict = sweep_range(attn_values, attn_obj, gw_obj, tag_handle, energy_pattern_val=energy_pattern,
                                     num_of_packets=num_of_packets)
        
        fine_attn_df = pd.DataFrame.from_dict(fine_attn_dict, orient='index')
        fine_attn_df.columns = ['num_of_packets', 'primary_tag_percentage', 'secondary_tag_percentage', 'primary_rssi',
                                'secondary_rssi', 'seconds_to_packets', 'primary_sprinkler_time',
                                'secondary_sprinkler_time']
        
        print('Final results for pattern: ', energy_pattern)
        if not fine_only:
            print('Coarse data: {attn_dict}'.format(attn_dict=coarse_attn_dict))
        print('Fine data: {attn_dict}'.format(attn_dict=fine_attn_dict))
    gw_obj.stop_continuous_listener()
    fine_attn_df = pd.DataFrame.from_dict(fine_attn_dict, orient='index')
    fine_attn_df.columns = ['num_of_packets', 'primary_tag_percentage', 'secondary_tag_percentage', 'primary_rssi',
                            'secondary_rssi', 'seconds_to_packets', 'primary_sprinkler_time',
                            'secondary_sprinkler_time']
    
    fine_attn_filtered_df = fine_attn_df[fine_attn_df['primary_sprinkler_time'] >= td_lower_limit]
    fine_attn_filtered_df = fine_attn_df.drop(fine_attn_df[fine_attn_df.primary_tag_percentage < 0.4].index)
    
    path = Path(os.path.dirname(__file__))
    fine_attn_filtered_df.to_csv('{parent_path}\\output.csv'.format(parent_path=path))
    
    try:
        fine_attn_filtered_df['primary_sprinkler_time_decending'] = fine_attn_filtered_df.apply(
            lambda row: -1 * row.primary_sprinkler_time, axis=1)
    except Exception as e:
        pass
    
    top_attn_by_chars = fine_attn_filtered_df.sort_values(
        ['primary_tag_percentage', 'primary_sprinkler_time_decending'], ascending=False)
    
    first_preference = top_attn_by_chars.iloc[0].name
    primary_rssi = top_attn_by_chars.loc[first_preference].primary_rssi
    
    print('Set ATTN to ', first_preference, 'dB for best performance')
    
    path = '{parent_path}\\config\\equipment_config.json'.format(parent_path=path.parent)
    
    key = "equipment"
    value = {'optimal_attenuation': int(first_preference), 'primary_rssi_mean': primary_rssi}
    update_json_field(path, key, value)
    
    if set_optimal_attn:
        set_attn_status = set_attn_power(attn_obj, first_preference)
    else:
        print('Setting ATTN to 0 dB')
        set_attn_status = set_attn_power(attn_obj, 0)


def set_calibration_attn(set_optimal_attn=True, config_path=None):
    """
    setting attn value
    :type set_optimal_attn: Bool
    :param set_optimal_attn: A flag to decide if set attn to optimal value
    :type config_path: string
    :param config_path: path to config file

    return:
        attenuation set to attenuator or None if set is failed
    """
    # create equipment objects
    try:
        attn_obj = Attenuator('API').GetActiveTE()
        current_attn = attn_obj.Getattn()
    except Exception as e:
        raise EquipmentError('Attenuator Error - Verify Attenuator connection')
    
    if config_path is None:
        parent_path = os.path.dirname(__file__)
        path = '{parent_path}\\config\\equipment_config.json'.format(parent_path=parent_path)
    else:
        path = config_path
    
    key = "equipment"
    optimal_attenuation_value = None
    if set_optimal_attn:
        try:
            with open(path, "r+") as jsonFile:
                config_data = json.load(jsonFile)
                optimal_attenuation_value = config_data.get(key).get('optimal_attenuation')
        except Exception as e:
            raise
    
    if set_optimal_attn and optimal_attenuation_value is not None:
        print('Setting ATTN to {value} dB'.format(value=optimal_attenuation_value))
        set_attn_status = set_attn_power(attn_obj, optimal_attenuation_value)
        if set_attn_status:
            return optimal_attenuation_value
    else:
        print('Setting ATTN to 0 dB')
        set_attn_status = set_attn_power(attn_obj, 0)
        if set_attn_status:
            return 0
    return None


# main
if __name__ == '__main__':
    start_calibration()
